import json
import logging
import multiprocessing as mp
import pickle
import time
from collections import Counter
from copy import deepcopy
from typing import List, Union, Tuple

import numpy as np
from flare.env import AtomicEnvironment
from flare.gp_algebra import (
    get_like_from_mats,
    get_neg_like_grad,
    get_ky_mat_update,
    _global_training_data,
    _global_training_labels,
    _global_training_structures,
    _global_energy_labels,
    get_Ky_mat,
    get_kernel_vector,
    en_kern_vec,
    efs_kern_vec,
)
from flare.kernels.utils import (
    str_to_kernel_set,
    from_mask_to_args,
    kernel_str_to_array,
)
from flare.output import Output, set_logger
from flare.parameters import Parameters
from flare.struc import Structure
from flare.utils.element_coder import NumpyEncoder, Z_to_element
from numpy.random import random
from scipy.linalg import solve_triangular
from scipy.optimize import minimize


class GaussianProcess:
    """Gaussian process force field. Implementation is based on Algorithm 2.1
    (pg. 19) of "Gaussian Processes for Machine Learning" by Rasmussen and
    Williams.

    Methods within GaussianProcess allow you to make predictions on
    AtomicEnvironment objects (see env.py) generated from
    FLARE Structures (see struc.py), and after data points are added,
    optimize hyperparameters based on available training data (train method).

    Args:
        kernels (list, optional): Determine the type of kernels. Example:
            ['twbody', 'threebody'], ['2', '3', 'mb'], ['2']. Defaults to [
            'twboody', 'threebody']
        component (str, optional): Determine single- ("sc") or multi-
            component ("mc") kernel to use. Defaults to "mc"
        hyps (np.ndarray, optional): Hyperparameters of the GP.
        cutoffs (Dict, optional): Cutoffs of the GP kernel. For simple hyper-
            parameter setups, formatted like {"twobody":7, "threebody":4.5},
            etc.
        hyp_labels (List, optional): List of hyperparameter labels. Defaults
            to None.
        opt_algorithm (str, optional): Hyperparameter optimization algorithm.
            Defaults to 'L-BFGS-B'.
        maxiter (int, optional): Maximum number of iterations of the
            hyperparameter optimization algorithm. Defaults to 10.
        parallel (bool, optional): If True, the covariance matrix K of the GP is
            computed in parallel. Defaults to False.
        n_cpus (int, optional): Number of cpus used for parallel
            calculations. Defaults to 1 (serial)
        n_sample (int, optional): Size of submatrix to use when parallelizing
            predictions.
        output (Output, optional): Output object used to dump hyperparameters
            during optimization. Defaults to None.
        hyps_mask (dict, optional): hyps_mask can set up which hyper parameter
            is used for what interaction. Details see kernels/mc_sephyps.py
        name (str, optional): Name for the GP instance which dictates global
            memory access.
    """

    def __init__(
        self,
        kernels: List[str] = None,
        component: str = "mc",
        hyps: "ndarray" = None,
        cutoffs: dict = None,
        hyps_mask: dict = None,
        hyp_labels: List = None,
        opt_algorithm: str = "L-BFGS-B",
        maxiter: int = 10,
        parallel: bool = False,
        per_atom_par: bool = True,
        n_cpus: int = 1,
        n_sample: int = 100,
        output: Output = None,
        name="default_gp",
        energy_noise: float = 0.01,
        **kwargs,
    ):
        """Initialize GP parameters and training data."""

        # load arguments into attributes
        self.name = name

        self.output = output
        self.opt_algorithm = opt_algorithm

        self.per_atom_par = per_atom_par
        self.maxiter = maxiter

        # set up parallelization
        self.n_cpus = n_cpus
        self.n_sample = n_sample
        self.parallel = parallel

        self.component = component
        self.kernels = (
            ["twobody", "threebody"]
            if kernels is None
            else kernel_str_to_array("".join(kernels))
        )
        self.cutoffs = {} if cutoffs is None else cutoffs
        self.hyp_labels = hyp_labels
        self.hyps_mask = {} if hyps_mask is None else hyps_mask
        self.hyps = hyps

        GaussianProcess.backward_arguments(kwargs, self.__dict__)
        GaussianProcess.backward_attributes(self.__dict__)

        # ------------  "computed" attributes ------------

        if self.output is None:
            self.logger_name = self.name + "GaussianProcess"
            set_logger(self.logger_name, stream=True, fileout_name=None, verbose="info")
        else:
            self.logger_name = self.output.basename + "log"

        if self.hyps is None:
            # If no hyperparameters are passed in, assume 2 hyps for each
            # kernel, plus one noise hyperparameter, and use a guess value
            self.hyps = np.array([0.1] * (1 + 2 * len(self.kernels)))
        else:
            self.hyps = np.array(self.hyps, dtype=np.float64)

        kernel, grad, ek, efk, efs_e, efs_f, efs_self = str_to_kernel_set(
            self.kernels, self.component, self.hyps_mask
        )
        self.kernel = kernel
        self.kernel_grad = grad
        self.energy_force_kernel = efk
        self.energy_kernel = ek
        self.efs_energy_kernel = efs_e
        self.efs_force_kernel = efs_f
        self.efs_self_kernel = efs_self
        self.kernels = kernel_str_to_array(kernel.__name__)

        # parallelization
        if self.parallel:
            if self.n_cpus is None:
                self.n_cpus = mp.cpu_count()
            else:
                self.n_cpus = n_cpus
        else:
            self.n_cpus = 1

        self.training_data = []  # Atomic environments
        self.training_labels = []  # Forces acting on central atoms
        self.training_labels_np = np.empty(
            0,
        )
        self.n_envs_prev = len(self.training_data)

        # Attributes to accomodate energy labels:
        self.training_structures = []  # Environments of each structure
        self.energy_labels = []  # Energies of training structures
        self.energy_labels_np = np.empty(
            0,
        )
        self.energy_noise = energy_noise
        self.all_labels = np.empty(
            0,
        )

        # Parameters set during training
        self.ky_mat = None
        self.force_block = None
        self.energy_block = None
        self.force_energy_block = None
        self.l_mat = None
        self.l_mat_inv = None
        self.alpha = None
        self.ky_mat_inv = None
        self.likelihood = None
        self.likelihood_gradient = None
        self.bounds = None

        # File used for reading / writing model if model is large
        self.ky_mat_file = None
        # Flag if too-big warning has been printed for this model
        self.large_warning = False

        if self.logger_name is None:
            if self.output is None:
                self.logger_name = self.name + "GaussianProcess"
                set_logger(
                    self.logger_name, stream=True, fileout_name=None, verbose="info"
                )
            else:
                self.logger_name = self.output.basename + "log"
        logger = logging.getLogger(self.logger_name)

        if self.cutoffs == {}:
            # If no cutoffs are passed in, assume 7 A for 2 body, 3.5 for
            # 3-body.
            cutoffs = {}
            if "twobody" in self.kernels:
                cutoffs["twobody"] = 7
            if "threebody" in self.kernels:
                cutoffs["threebody"] = 3.5
            if "manybody" in self.kernels:
                raise ValueError(
                    "No cutoff was set for the manybody kernel."
                    "A default value will not be set by default."
                )

            self.cutoffs = cutoffs
            logger.warning(
                "Warning: No cutoffs were set for your GP."
                "Default values have been assigned but you "
                "should think carefully about which are "
                "appropriate for your use case."
            )

        self.check_instantiation()

    @property
    def force_noise(self):
        return Parameters.get_noise(self.hyps_mask, self.hyps, constraint=False)

    @property
    def hyps_and_labels(self):
        return Parameters.get_hyps(
            self.hyps_mask, self.hyps, constraint=False, label=True
        )

    def check_instantiation(self):
        """
        Runs a series of checks to ensure that the user has not supplied
        contradictory arguments which will result in undefined behavior
        with multiple hyperparameters.
        :return:
        """
        logger = logging.getLogger(self.logger_name)

        # check whether it's be loaded before
        loaded = False
        if self.name in _global_training_labels:
            if (
                _global_training_labels.get(self.name, None)
                is not self.training_labels_np
            ):
                loaded = True
        if self.name in _global_energy_labels:
            if _global_energy_labels.get(self.name, None) is not self.energy_labels_np:
                loaded = True

        if loaded:

            base = f"{self.name}"
            count = 2
            while self.name in _global_training_labels and count < 100:
                time.sleep(random())
                self.name = f"{base}_{count}"
                logger.debug(
                    "Specified GP name is present in global memory; "
                    "Attempting to rename the "
                    f"GP instance to {self.name}"
                )
                count += 1
            if self.name in _global_training_labels:
                milliseconds = int(round(time.time() * 1000) % 10000000)
                self.name = f"{base}_{milliseconds}"
                logger.debug(
                    "Specified GP name still present in global memory: "
                    f"renaming the gp instance to {self.name}"
                )
            logger.debug(f"Final name of the gp instance is {self.name}")

        self.sync_data()

        self.hyps_mask = Parameters.check_instantiation(
            hyps=self.hyps,
            cutoffs=self.cutoffs,
            kernels=self.kernels,
            param_dict=self.hyps_mask,
        )

        self.bounds = deepcopy(self.hyps_mask.get("bounds", None))

    def update_kernel(
        self,
        kernels: List[str],
        component: str = "mc",
        hyps=None,
        cutoffs: dict = None,
        hyps_mask: dict = None,
    ):
        kernel, grad, ek, efk, _, _, _ = str_to_kernel_set(
            kernels, component, hyps_mask
        )
        self.kernel = kernel
        self.kernel_grad = grad
        self.energy_force_kernel = efk
        self.energy_kernel = ek
        self.kernels = kernel_str_to_array(kernel.__name__)

        if hyps_mask is not None:
            self.hyps_mask = hyps_mask
        # Cutoffs argument will override hyps mask's cutoffs key, if present
        if isinstance(hyps_mask, dict) and cutoffs is None:
            cutoffs = hyps_mask.get("cutoffs", None)

        if cutoffs is not None:
            if self.cutoffs != cutoffs:
                self.adjust_cutoffs(cutoffs, train=False, new_hyps_mask=hyps_mask)
            self.cutoffs = cutoffs

        if isinstance(hyps_mask, dict) and hyps is None:
            hyps = hyps_mask.get("hyps", None)

        if hyps is not None:
            self.hyps = hyps

    def update_db(
        self,
        struc: Structure,
        forces: "ndarray" = None,
        custom_range: List[int] = (),
        energy: float = None,
        stress: "ndarray" = None,
    ):
        """Given a structure and forces, add local environments from the
        structure to the training set of the GP. If energy is given, add the
        entire structure to the training set.

        Args:
            struc (Structure): Input structure. Local environments of atoms
                in this structure will be added to the training set of the GP.

            forces (np.ndarray): Forces on atoms in the structure.

            custom_range (List[int]): Indices of atoms whose local
                environments will be added to the training set of the GP.

            energy (float): Energy of the structure.

            stress (np.ndarray): Stress tensor of the structure. The stress
                tensor components should be given in the following order:
                xx, xy, xz, yy, yz, zz.
        """

        # By default, use all atoms in the structure
        noa = len(struc.positions)
        update_indices = custom_range or list(range(noa))

        # If forces are given, update the environment list.
        if forces is not None:
            for atom in update_indices:
                env_curr = AtomicEnvironment(
                    struc, atom, self.cutoffs, cutoffs_mask=self.hyps_mask
                )
                forces_curr = np.array(forces[atom])

                self.training_data.append(env_curr)
                self.training_labels.append(forces_curr)

            # create numpy array of training labels
            self.training_labels_np = np.hstack(self.training_labels)

        # If an energy is given, update the structure list.
        if energy is not None:
            structure_list = []  # Populate with all environments of the struc
            for atom in range(noa):
                env_curr = AtomicEnvironment(
                    struc, atom, self.cutoffs, cutoffs_mask=self.hyps_mask
                )
                structure_list.append(env_curr)

            self.energy_labels.append(energy)
            self.training_structures.append(structure_list)
            self.energy_labels_np = np.array(self.energy_labels)

        if forces is None and energy is None and stress is None:
            logger = logging.getLogger(self.logger_name)
            logger.warn(
                "Update DB method called with data but no labels!"
                "The GP has not been updated with data!"
            )
        # update list of all labels
        self.all_labels = np.concatenate(
            (self.training_labels_np, self.energy_labels_np)
        )
        self.sync_data()

    def add_one_env(
        self,
        env: AtomicEnvironment,
        force: "np.ndarray" = None,
        train: bool = False,
        **kwargs,
    ):
        """Add a single local environment to the training set of the GP.

        Args:
            env (AtomicEnvironment): Local environment to be added to the
                training set of the GP.
            force (np.ndarray): Force on the central atom of the local
                environment in the form of a 3-component Numpy array
                containing the x, y, and z components.
            train (bool): If True, the GP is trained after the local
                environment is added.
        """
        self.training_data.append(env)
        if force is None:
            self.training_labels.append(env.force)
        else:
            self.training_labels.append(force)
        self.training_labels_np = np.hstack(self.training_labels)
        self.sync_data()

        # update list of all labels
        self.all_labels = np.concatenate(
            (self.training_labels_np, self.energy_labels_np)
        )

        if train:
            self.train(**kwargs)

    def train(
        self,
        logger_name: str = None,
        custom_bounds=None,
        grad_tol: float = 1e-4,
        x_tol: float = 1e-5,
        line_steps: int = 20,
        print_progress: bool = False,
    ):
        """Train Gaussian Process model on training data. Tunes the
        hyperparameters to maximize the likelihood, then computes L and alpha
        (related to the covariance matrix of the training set).

        Args:
            logger (logging.logger): logger object specifying where to write the
                progress of the optimization.
            custom_bounds (np.ndarray): Custom bounds on the hyperparameters.
            grad_tol (float): Tolerance of the hyperparameter gradient that
                determines when hyperparameter optimization is terminated.
            x_tol (float): Tolerance on the x values used to decide when
                Nelder-Mead hyperparameter optimization is terminated.
            line_steps (int): Maximum number of line steps for L-BFGS
                hyperparameter optimization.
                :param logger_name:
                :param print_progress:
        """

        verbose = "warning"
        if print_progress:
            verbose = "info"
        if logger_name is None:
            set_logger(
                "gp_algebra",
                stream=True,
                fileout_name="log.gp_algebra",
                verbose=verbose,
            )
            logger_name = "gp_algebra"

        disp = print_progress

        if (
            max(len(self.training_data), len(self.training_labels)) > 5000
            and not self.large_warning
        ):

            self.large_warning = True
            warning_message = (
                "WARNING! Your GP is very large (>5000 atomic "
                "environments). The hyperparameter optimization process "
                "does not scale favorably with increasing atomic "
                "environments"
                " (roughly N^2)"
                "and so your GP may take a very long time to train."
                "Consider finding a way to reduce the number of atomic "
                "environments in your model if you want to optimize the "
                "hyperparameters or optimize them by a different route."
            )

            logger = logging.getLogger(self.logger_name)
            logger.warning(warning_message)

        if len(self.training_data) == 0 or len(self.training_labels) == 0:
            raise Warning(
                "You are attempting to train a GP with no "
                "training data. Add environments and forces "
                "to the GP and try again."
            )

        x_0 = self.hyps

        args = (
            self.name,
            self.kernel_grad,
            logger_name,
            self.cutoffs,
            self.hyps_mask,
            self.n_cpus,
            self.n_sample,
        )

        res = None

        if self.opt_algorithm == "L-BFGS-B":

            # bound signal noise below to avoid overfitting
            if self.bounds is None:
                bounds = np.array([(1e-6, np.inf)] * len(x_0))
                bounds[-1, 0] = 1e-3
            else:
                bounds = self.bounds

            # Catch linear algebra errors and switch to BFGS if necessary
            try:
                res = minimize(
                    get_neg_like_grad,
                    x_0,
                    args,
                    method="L-BFGS-B",
                    jac=True,
                    bounds=bounds,
                    options={
                        "disp": disp,
                        "gtol": grad_tol,
                        "maxls": line_steps,
                        "maxiter": self.maxiter,
                    },
                )
            except np.linalg.LinAlgError:
                logger = logging.getLogger(self.logger_name)
                logger.warning(
                    "Algorithm for L-BFGS-B failed. Changing to "
                    "BFGS for remainder of run."
                )
                self.opt_algorithm = "BFGS"

        if custom_bounds is not None:
            res = minimize(
                get_neg_like_grad,
                x_0,
                args,
                method="L-BFGS-B",
                jac=True,
                bounds=custom_bounds,
                options={
                    "disp": disp,
                    "gtol": grad_tol,
                    "maxls": line_steps,
                    "maxiter": self.maxiter,
                },
            )

        elif self.opt_algorithm == "BFGS":
            res = minimize(
                get_neg_like_grad,
                x_0,
                args,
                method="BFGS",
                jac=True,
                options={"disp": disp, "gtol": grad_tol, "maxiter": self.maxiter},
            )

        if res is None:
            raise RuntimeError("Optimization failed for some reason.")
        self.hyps = res.x
        self.set_L_alpha()
        self.likelihood = -res.fun
        self.likelihood_gradient = -res.jac

        return res

    def check_L_alpha(self):
        """
        Check that the alpha vector is up to date with the training set. If
        not, update_L_alpha is called.
        """

        # Check that alpha is up to date with training set
        size3 = len(self.training_data) * 3 + len(self.training_structures)

        # If model is empty, then just return
        if size3 == 0:
            return

        if self.alpha is None:
            self.update_L_alpha()
        elif size3 > self.alpha.shape[0]:
            self.update_L_alpha()
        elif size3 != self.alpha.shape[0]:
            self.set_L_alpha()

    def predict(self, x_t: AtomicEnvironment, d: int) -> [float, float]:
        """
        Predict a force component of the central atom of a local environment.

        Args:
            x_t (AtomicEnvironment): Input local environment.
            d (int): Force component to be predicted (1 is x, 2 is y, and
                3 is z).

        Return:
            (float, float): Mean and epistemic variance of the prediction.
        """

        if d not in [1, 2, 3]:
            raise ValueError("d should be 1, 2, or 3")

        # Kernel vector allows for evaluation of atomic environments.
        if self.parallel and not self.per_atom_par:
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        self.sync_data()

        k_v = get_kernel_vector(
            self.name,
            self.kernel,
            self.energy_force_kernel,
            x_t,
            d,
            self.hyps,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=n_cpus,
            n_sample=self.n_sample,
        )

        # Guarantee that alpha is up to date with training set
        self.check_L_alpha()

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance without cholesky (possibly faster)
        # pass args to kernel based on if mult. hyperparameters in use
        args = from_mask_to_args(self.hyps, self.cutoffs, self.hyps_mask)

        self_kern = self.kernel(x_t, x_t, d, d, *args)
        pred_var = self_kern - np.matmul(np.matmul(k_v, self.ky_mat_inv), k_v)

        return pred_mean, pred_var

    def predict_force_xyz(self, x_t: AtomicEnvironment) -> ("np.ndarray", "np.ndarray"):
        """
        Simple wrapper to predict all three components of a force in one go.
        :param x_t:
        :return:
        """
        forces = []
        stds = []
        for d in (1, 2, 3):
            force, std = self.predict(x_t, d)
            forces.append(force)
            stds.append(std)
        return np.array(forces), np.array(stds)

    def predict_local_energy(self, x_t: AtomicEnvironment) -> float:
        """Predict the local energy of a local environment.

        Args:
            x_t (AtomicEnvironment): Input local environment.

        Return:
            float: Local energy predicted by the GP.
        """

        if self.parallel and not self.per_atom_par:
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        self.sync_data()

        k_v = en_kern_vec(
            self.name,
            self.energy_force_kernel,
            self.energy_kernel,
            x_t,
            self.hyps,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=n_cpus,
            n_sample=self.n_sample,
        )

        pred_mean = np.matmul(k_v, self.alpha)

        return pred_mean

    def predict_local_energy_and_var(self, x_t: AtomicEnvironment):
        """Predict the local energy of a local environment and its
        uncertainty.

        Args:
            x_t (AtomicEnvironment): Input local environment.

        Return:
            (float, float): Mean and predictive variance predicted by the GP.
        """

        if self.parallel and not self.per_atom_par:
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        self.sync_data()

        # get kernel vector
        k_v = en_kern_vec(
            self.name,
            self.energy_force_kernel,
            self.energy_kernel,
            x_t,
            self.hyps,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=n_cpus,
            n_sample=self.n_sample,
        )

        # get predictive mean
        pred_mean = np.matmul(k_v, self.alpha)

        # get predictive variance
        v_vec = solve_triangular(self.l_mat, k_v, lower=True)
        args = from_mask_to_args(self.hyps, self.cutoffs, self.hyps_mask)

        self_kern = self.energy_kernel(x_t, x_t, *args)

        pred_var = self_kern - np.matmul(v_vec, v_vec)

        return pred_mean, pred_var

    def predict_efs(self, x_t: AtomicEnvironment):
        """Predict the local energy, forces, and partial stresses of an
        atomic environment and their predictive variances."""

        # Kernel vector allows for evaluation of atomic environments.
        if self.parallel and not self.per_atom_par:
            n_cpus = self.n_cpus
        else:
            n_cpus = 1

        self.sync_data()

        energy_vector, force_array, stress_array = efs_kern_vec(
            self.name,
            self.efs_force_kernel,
            self.efs_energy_kernel,
            x_t,
            self.hyps,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=n_cpus,
            n_sample=self.n_sample,
        )

        # Check that alpha is up to date with training set.
        self.check_L_alpha()

        # Compute mean predictions.
        en_pred = np.matmul(energy_vector, self.alpha)
        force_pred = np.matmul(force_array, self.alpha)
        stress_pred = np.matmul(stress_array, self.alpha)

        # Compute uncertainties.
        args = from_mask_to_args(self.hyps, self.cutoffs, self.hyps_mask)
        self_en, self_force, self_stress = self.efs_self_kernel(x_t, *args)

        en_var = self_en - np.matmul(
            np.matmul(energy_vector, self.ky_mat_inv), energy_vector
        )
        force_var = self_force - np.diag(
            np.matmul(np.matmul(force_array, self.ky_mat_inv), force_array.transpose())
        )
        stress_var = self_stress - np.diag(
            np.matmul(
                np.matmul(stress_array, self.ky_mat_inv), stress_array.transpose()
            )
        )

        return en_pred, force_pred, stress_pred, en_var, force_var, stress_var

    def set_L_alpha(self):
        """
        Invert the covariance matrix, setting L (a lower triangular
        matrix s.t. L L^T = (K + sig_n^2 I)) and alpha, the inverse
        covariance matrix multiplied by the vector of training labels.
        The forces and variances are later obtained using alpha.
        """

        self.sync_data()

        ky_mat = get_Ky_mat(
            self.hyps,
            self.name,
            self.kernel,
            self.energy_kernel,
            self.energy_force_kernel,
            self.energy_noise,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=self.n_cpus,
            n_sample=self.n_sample,
        )

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.all_labels)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv

        self.likelihood = get_like_from_mats(ky_mat, l_mat, alpha, self.name)
        self.n_envs_prev = len(self.training_data)

    def update_L_alpha(self):
        """
        Update the GP's L matrix and alpha vector without recalculating
        the entire covariance matrix K.
        """

        # Set L matrix and alpha if set_L_alpha has not been called yet
        if self.l_mat is None or np.array(self.ky_mat) is np.array(None):
            self.set_L_alpha()
            return

        # Reset global variables.
        self.sync_data()

        ky_mat = get_ky_mat_update(
            self.ky_mat,
            self.n_envs_prev,
            self.hyps,
            self.name,
            self.kernel,
            self.energy_kernel,
            self.energy_force_kernel,
            self.energy_noise,
            cutoffs=self.cutoffs,
            hyps_mask=self.hyps_mask,
            n_cpus=self.n_cpus,
            n_sample=self.n_sample,
        )

        l_mat = np.linalg.cholesky(ky_mat)
        l_mat_inv = np.linalg.inv(l_mat)
        ky_mat_inv = l_mat_inv.T @ l_mat_inv
        alpha = np.matmul(ky_mat_inv, self.all_labels)

        self.ky_mat = ky_mat
        self.l_mat = l_mat
        self.alpha = alpha
        self.ky_mat_inv = ky_mat_inv
        self.n_envs_prev = len(self.training_data)

    def __str__(self):
        """String representation of the GP model."""

        thestr = ""
        thestr += f"Number of cpu cores: {self.n_cpus}\n"
        thestr += f"Kernel: {self.kernels}\n"
        thestr += f"Training points: {len(self.training_data)}\n"
        thestr += f"Cutoffs: {self.cutoffs}\n"

        thestr += f"Number of hyperparameters: {len(self.hyps)}\n"
        thestr += f"Hyperparameter array: {str(self.hyps)}\n"

        if self.hyp_labels is None:
            # Put unlabeled hyperparameters on one line
            thestr = thestr[:-1]
            thestr += str(self.hyps) + "\n"
        else:
            for hyp, label in zip(self.hyps, self.hyp_labels):
                thestr += f"{label}: {hyp} \n"

        return thestr

    def as_dict(self):
        """Dictionary representation of the GP model."""

        self.check_L_alpha()

        out_dict = dict(vars(self))

        out_dict["training_data"] = [env.as_dict() for env in self.training_data]

        # Write training structures (which are just list of environments)
        out_dict["training_structures"] = []
        for n, env_list in enumerate(self.training_structures):
            out_dict["training_structures"].append([])
            for env_curr in env_list:
                out_dict["training_structures"][n].append(env_curr.as_dict())

        # Remove the callables
        for key in [
            "kernel",
            "kernel_grad",
            "energy_kernel",
            "energy_force_kernel",
            "efs_energy_kernel",
            "efs_force_kernel",
            "efs_self_kernel",
            "output",
        ]:
            out_dict.pop(key)

        return out_dict

    def sync_data(self):
        _global_training_data[self.name] = self.training_data
        _global_training_labels[self.name] = self.training_labels_np
        _global_training_structures[self.name] = self.training_structures
        _global_energy_labels[self.name] = self.energy_labels_np

    @staticmethod
    def from_dict(dictionary):
        """Create GP object from dictionary representation."""

        GaussianProcess.backward_arguments(dictionary, dictionary)
        GaussianProcess.backward_attributes(dictionary)

        new_gp = GaussianProcess(**dictionary)

        # Save time by attempting to load in computed attributes
        if "training_data" in dictionary:
            new_gp.training_data = [
                AtomicEnvironment.from_dict(env) for env in dictionary["training_data"]
            ]
            new_gp.training_labels = deepcopy(dictionary["training_labels"])
            new_gp.training_labels_np = deepcopy(dictionary["training_labels_np"])
            new_gp.sync_data()

        # Reconstruct training structures.
        if "training_structures" in dictionary:
            new_gp.training_structures = []
            for n, env_list in enumerate(dictionary["training_structures"]):
                new_gp.training_structures.append([])
                for env_curr in env_list:
                    new_gp.training_structures[n].append(
                        AtomicEnvironment.from_dict(env_curr)
                    )
            new_gp.energy_labels = deepcopy(dictionary["energy_labels"])
            new_gp.energy_labels_np = deepcopy(dictionary["energy_labels_np"])
            new_gp.sync_data()

        new_gp.all_labels = np.concatenate(
            (new_gp.training_labels_np, new_gp.energy_labels_np)
        )

        new_gp.likelihood = dictionary.get("likelihood", None)
        new_gp.likelihood_gradient = dictionary.get("likelihood_gradient", None)

        new_gp.n_envs_prev = len(new_gp.training_data)

        # Save time by attempting to load in computed attributes
        if dictionary.get("ky_mat_file"):
            try:
                new_gp.ky_mat = np.load(dictionary["ky_mat_file"])
                new_gp.compute_matrices()
                new_gp.ky_mat_file = None

            except FileNotFoundError:
                new_gp.ky_mat = None
                new_gp.l_mat = None
                new_gp.alpha = None
                new_gp.ky_mat_inv = None
                filename = dictionary.get("ky_mat_file")
                logger = logging.getLogger(new_gp.logger_name)
                logger.warning(
                    "the covariance matrices are not loaded"
                    f"because {filename} cannot be found"
                )
        else:
            new_gp.ky_mat = (
                np.array(dictionary["ky_mat"])
                if dictionary.get("ky_mat") is not None
                else None
            )
            new_gp.ky_mat_inv = (
                np.array(dictionary["ky_mat_inv"])
                if dictionary.get("ky_mat_inv") is not None
                else None
            )
            new_gp.ky_mat = (
                np.array(dictionary["ky_mat"])
                if dictionary.get("ky_mat") is not None
                else None
            )
            new_gp.l_mat = (
                np.array(dictionary["l_mat"])
                if dictionary.get("l_mat") is not None
                else None
            )
            new_gp.alpha = (
                np.array(dictionary["alpha"])
                if dictionary.get("alpha") is not None
                else None
            )

        return new_gp

    def compute_matrices(self):
        """
        When covariance matrix is known, reconstruct other matrices.
        Used in re-loading large GPs.
        :return:
        """
        ky_mat = self.ky_mat

        if ky_mat is None or (isinstance(ky_mat, np.ndarray) and not np.any(ky_mat)):
            Warning(
                "Warning: Covariance matrix was not loaded but "
                "compute_matrices was called. Computing covariance "
                "matrix and proceeding..."
            )
            self.set_L_alpha()

        else:
            self.l_mat = np.linalg.cholesky(ky_mat)
            self.l_mat_inv = np.linalg.inv(self.l_mat)
            self.ky_mat_inv = self.l_mat_inv.T @ self.l_mat_inv
            self.alpha = np.matmul(self.ky_mat_inv, self.all_labels)

    def adjust_cutoffs(
        self,
        new_cutoffs: Union[list, tuple, "np.ndarray"] = None,
        reset_L_alpha=True,
        train=True,
        new_hyps_mask=None,
    ):
        """
        Loop through atomic environment objects stored in the training data,
        and re-compute cutoffs for each. Useful if you want to gauge the
        impact of cutoffs given a certain training set! Unless you know
        *exactly* what you are doing for some development or test purpose,
        it is **highly** suggested that you call set_L_alpha and
        re-optimize your hyperparameters afterwards as is default here.

        A helpful way to update the cutoffs and kernel for an extant
        GP is to perform the following commands:
        >> hyps_mask = pm.as_dict()
        >> hyps = hyps_mask['hyps']
        >> cutoffs = hyps_mask['cutoffs']
        >> kernels = hyps_mask['kernels']
        >> gp_model.update_kernel(kernels, 'mc', hyps, cutoffs, hyps_mask)

        :param reset_L_alpha:
        :param train:
        :param new_hyps_mask:
        :param new_cutoffs:
        :return:
        """

        if new_hyps_mask is not None:
            hm = new_hyps_mask
            self.hyps_mask = new_hyps_mask
        else:
            hm = self.hyps_mask
        if new_cutoffs is None:
            try:
                new_cutoffs = hm["cutoffs"]
            except KeyError:
                raise KeyError(
                    "New cutoffs not found in the hyps_mask"
                    "dictionary via call to 'cutoffs' key."
                )

        # update environment
        nenv = len(self.training_data)
        for i in range(nenv):
            self.training_data[i].cutoffs = new_cutoffs
            self.training_data[i].cutoffs_mask = hm
            self.training_data[i].setup_mask(hm)
            self.training_data[i].compute_env()

        # Ensure that training data and labels are still consistent
        self.sync_data()

        self.cutoffs = new_cutoffs

        if reset_L_alpha:
            del self.l_mat
            del self.ky_mat
            self.set_L_alpha()

        if train:
            self.train()

    def remove_force_data(
        self, indexes: Union[int, List[int]], update_matrices: bool = True
    ) -> Tuple[List[Structure], List["ndarray"]]:
        """
        Remove force components from the model. Convenience function which
        deletes individual data points.

        Matrices should *always* be updated if you intend to use the GP to make
        predictions afterwards. This might be time consuming for large GPs,
        so, it is provided as an option, but, only do so with extreme caution.
        (Undefined behavior may result if you try to make predictions and/or
        add to the training set afterwards).

        Returns training data which was removed akin to a pop method, in order
        of lowest to highest index passed in.

        :param indexes: Indexes of envs in training data to remove.
        :param update_matrices: If false, will not update the GP's matrices
            afterwards (which can be time consuming for large models).
            This should essentially always be true except for niche development
            applications.
        :return:
        """

        # Listify input even if one integer
        if isinstance(indexes, int):
            indexes = [indexes]

        if max(indexes) > len(self.training_data):
            raise ValueError("Index out of range of data")

        if len(indexes) == 0:
            return [], []

        # Get in reverse order so that modifying higher indexes doesn't affect
        # lower indexes
        indexes.sort(reverse=True)
        removed_data = []
        removed_labels = []
        for i in indexes:
            removed_data.append(self.training_data.pop(i))
            removed_labels.append(self.training_labels.pop(i))

        self.training_labels_np = np.hstack(self.training_labels)
        self.all_labels = np.concatenate(
            (self.training_labels_np, self.energy_labels_np)
        )
        self.sync_data()

        if update_matrices:
            self.set_L_alpha()

        # Put removed data in order of lowest to highest index
        removed_data.reverse()
        removed_labels.reverse()

        return removed_data, removed_labels

    def write_model(
        self, name: str, format: str = None, split_matrix_size_cutoff: int = 5000
    ):
        """
        Write model in a variety of formats to a file for later re-use.
        JSON files are open to visual inspection and are easier to use
        across different versions of FLARE or GP implementations. However,
        they are larger and loading them in takes longer (by setting up a
        new GP from the specifications). Pickled files can be faster to
        read & write, and they take up less memory.

        Args:
            name (str): Output name.
            format (str): Output format.
            split_matrix_size_cutoff (int): If there are more than this
            number of training points in the set, save the matrices seperately.
        """

        if len(self.training_data) > split_matrix_size_cutoff:
            np.save(f"{name}_ky_mat.npy", self.ky_mat)
            self.ky_mat_file = f"{name}_ky_mat.npy"

            temp_ky_mat = self.ky_mat
            temp_l_mat = self.l_mat
            temp_alpha = self.alpha
            temp_ky_mat_inv = self.ky_mat_inv

            self.ky_mat = None
            self.l_mat = None
            self.alpha = None
            self.ky_mat_inv = None

        # Automatically detect output format from name variable

        for detect in ["json", "pickle", "binary"]:
            if detect in name.lower():
                format = detect
                break

        if format is None:
            format = "json"

        supported_formats = ["json", "pickle", "binary"]

        if format.lower() == "json":
            if ".json" != name[-5:]:
                name += ".json"
            with open(name, "w") as f:
                json.dump(self.as_dict(), f, cls=NumpyEncoder)

        elif format.lower() == "pickle" or format.lower() == "binary":
            if ".pickle" != name[-7:]:
                name += ".pickle"
            with open(name, "wb") as f:
                pickle.dump(self, f)

        else:
            raise ValueError(
                "Output format not supported: try from {}".format(supported_formats)
            )

        if len(self.training_data) > split_matrix_size_cutoff:
            self.ky_mat = temp_ky_mat
            self.l_mat = temp_l_mat
            self.alpha = temp_alpha
            self.ky_mat_inv = temp_ky_mat_inv

    @staticmethod
    def from_file(filename: str, format: str = ""):
        """
        One-line convenience method to load a GP from a file stored using
        write_file

        Args:
            filename (str): path to GP model
            format (str): json or pickle if format is not in filename
        :return:
        """

        if ".json" in filename or "json" in format:
            with open(filename, "r") as f:
                gp_model = GaussianProcess.from_dict(json.loads(f.readline()))

        elif ".pickle" in filename or "pickle" in format:
            with open(filename, "rb") as f:

                gp_model = pickle.load(f)

                GaussianProcess.backward_arguments(gp_model.__dict__, gp_model.__dict__)

                GaussianProcess.backward_attributes(gp_model.__dict__)

                if hasattr(gp_model, "ky_mat_file") and gp_model.ky_mat_file:
                    try:
                        gp_model.ky_mat = np.load(
                            gp_model.ky_mat_file, allow_pickle=True
                        )
                        gp_model.compute_matrices()
                    except FileNotFoundError:
                        gp_model.ky_mat = None
                        gp_model.l_mat = None
                        gp_model.alpha = None
                        gp_model.ky_mat_inv = None
                        Warning(
                            f"the covariance matrices are not loaded, "
                            f"this can take a long time to recompute"
                        )

        else:
            raise ValueError(
                "Warning: Format unspecieified or file is not .json or .pickle format."
            )

        gp_model.check_instantiation()
        return gp_model

    def __len__(self):
        return len(self.training_data)

    @property
    def training_statistics(self) -> dict:
        """
        Return a dictionary with statistics about the current training data.
        Useful for quickly summarizing info about the GP.
        :return:
        """

        data = dict()

        data["N"] = len(self.training_data)

        # Count all of the present species in the atomic env. data
        present_species = []
        for env, _ in zip(self.training_data, self.training_labels):
            present_species.append(Z_to_element(env.structure.coded_species[env.atom]))

        # Summarize the relevant information
        data["species"] = list(set(present_species))
        data["envs_by_species"] = dict(Counter(present_species))

        return data

    @property
    def par(self):
        """
        Backwards compability attribute
        :return:
        """
        return self.parallel

    def __deepcopy__(self, memo):
        # this way can also deepcopy the training data in _global_training dicts
        return GaussianProcess.from_dict(self.as_dict())

    def __del__(self):
        if self is None:
            return
        if self.name in _global_training_labels:
            return (
                _global_training_data.pop(self.name, None),
                _global_training_labels.pop(self.name, None),
                _global_training_structures.pop(self.name, None),
                _global_energy_labels.pop(self.name, None),
            )

    @staticmethod
    def backward_arguments(kwargs, new_args={}):
        """
        update the initialize arguments that were renamed
        """

        if "kernel_name" in kwargs:
            DeprecationWarning("kernel_name is being replaced with kernels")
            new_args["kernels"] = kernel_str_to_array(kwargs["kernel_name"])
            kwargs.pop("kernel_name")
        if "nsample" in kwargs:
            DeprecationWarning("nsample is being replaced with n_sample")
            new_args["n_sample"] = kwargs["nsample"]
            kwargs.pop("nsample")
        if "par" in kwargs:
            DeprecationWarning("par is being replaced with parallel")
            new_args["parallel"] = kwargs["par"]
            kwargs.pop("par")
        if "no_cpus" in kwargs:
            DeprecationWarning("no_cpus is being replaced with n_cpu")
            new_args["n_cpus"] = kwargs["no_cpus"]
            kwargs.pop("no_cpus")
        if "multihyps" in kwargs:
            DeprecationWarning("multihyps is removed")
            kwargs.pop("multihyps")

        return new_args

    @staticmethod
    def backward_attributes(dictionary):
        """
        add new attributes to old instance
        or update attribute types
        """

        if "name" not in dictionary:
            dictionary["name"] = "default_gp"
        if "per_atom_par" not in dictionary:
            dictionary["per_atom_par"] = True
        if "opt_algorithm" not in dictionary:
            dictionary["opt_algorithm"] = "L-BFGS-B"
        if "hyps_mask" not in dictionary:
            dictionary["hyps_mask"] = None
        if "parallel" not in dictionary:
            dictionary["parallel"] = False
        if "component" not in dictionary:
            dictionary["component"] = "mc"

        if "training_structures" not in dictionary:
            # Environments of each structure
            dictionary["training_structures"] = []
            dictionary["energy_labels"] = []  # Energies of training structures
            dictionary["energy_labels_np"] = np.empty(
                0,
            )

        if "training_labels" not in dictionary:
            dictionary["training_labels"] = []
            dictionary["training_labels_np"] = np.empty(
                0,
            )

        if "energy_noise" not in dictionary:
            dictionary["energy_noise"] = 0.01

        if not isinstance(dictionary["cutoffs"], dict):
            dictionary["cutoffs"] = Parameters.cutoff_array_to_dict(
                dictionary["cutoffs"]
            )

        dictionary["hyps_mask"] = Parameters.backward(
            dictionary["kernels"], deepcopy(dictionary["hyps_mask"])
        )

        if "logger_name" not in dictionary:
            dictionary["logger_name"] = None



"""
Tool to enable the development of a GP model based on an AIMD
trajectory with many customizable options for fine control of training.
Contains methods to transfer the model to an OTF run or MD engine run.
"""
import json as json
import logging
import numpy as np
import time
import random as random
import warnings

from copy import deepcopy
from math import inf
from typing import List, Tuple, Union, Iterable, Dict

from flare.env import AtomicEnvironment
from flare.gp import GaussianProcess
from flare.mgp.mgp import MappedGaussianProcess
from flare.output import Output
from flare.predict import (
    predict_on_atom,
    predict_on_atom_en,
    predict_on_structure_par,
    predict_on_structure_par_en,
    predict_on_structure_mgp,
)
from flare.struc import Structure
from flare.utils.element_coder import element_to_Z, Z_to_element, NumpyEncoder
from flare.utils.learner import (
    subset_of_frame_by_element,
    is_std_in_bound_per_species,
    is_force_in_bound_per_species,
)
from flare.mgp import MappedGaussianProcess
from flare.parameters import Parameters
from flare.output import Output


class LearningProtocol(object):
    def __init__(
        self,
        model: Union[GaussianProcess, MappedGaussianProcess],
        n_cpus: int = 0,
        protocol_name: str = "learning",
    ):

        self.model = model
        self.output = Output(basename=protocol_name, always_flush=True)

    def train_model(self, **train_kwargs):

        if isinstance(self.model, GaussianProcess):
            self.model.train(train_kwargs)

        elif isinstance(self.model, MappedGaussianProcess):
            raise NotImplementedError("MGP training not defined")

        raise NotImplementedError("Must be implemented in child class")

    def augment_training_set(
        self,
        structure: Structure,
        selective_atoms: List[int] = list(),
        forces: Iterable[float] = None,
        energy: float = None,
    ):

        if forces is None:
            forces = structure.forces
        if energy is None and forces is None:
            energy = structure.energy

        if isinstance(self.model, GaussianProcess):

            self.model.update_db(
                structure=structure,
                selective_atoms=selective_atoms,
                forces=forces,
                energy=energy,
            )

        raise NotImplementedError("Must be implemented in child class")

    def get_next_structure(self):
        raise NotImplementedError("Must be implemented in child class")

    def check_prediction(self):
        raise NotImplementedError("Must be implemented in child class")

    def run_active_learning(self, init_frames: List[Structure]):
        """

        :param init_frames:
        :return:
        """
        raise NotImplementedError("Must be implemented in child class")

    def pre_run(self):

        if isinstance(self.model, GaussianProcess):
            self.output.write_header(
                str(self.gp),
                dt=0,
                Nsteps=len(self.frames),
                structure=None,
                optional={
                    "GP Statistics": json.dumps(self.gp.training_statistics),
                    "GP Name": self.gp.name,
                    "GP Write Name": self.output_name + "_model." + self.model_format,
                },
            )

    def run_passive_learning(
        self,
        frames: List[Structure] = list(),
        envs: List[AtomicEnvironment] = list(),
        max_atoms_from_frame: int = inf,
        shuffle_frames: bool = False,
        max_atoms_added: int = 1000,
        max_from_frame_by_species: Dict[str, int] = dict(),
    ):

        if not isinstance(self.model, GaussianProcess):
            raise NotImplementedError("Passive learning notyet configured for MGP")

        self.start_time = time.time()
        logger = logging.getLogger(self.logger_name)
        logger.debug("Now beginning passive learning.")

        # If seed environments were passed in, add them to the GP.

        for env in envs:
            self.model.add_one_env(env[0], env[1], train=False)
            logger.info("Added {str(env)} to dataset.")

        if shuffle_frames:
            frames = random.sample(frames, len(frames))

        n_atoms_added = 0

        for frame in frames:
            train_atoms = []
            # Get train atoms by species
            for species_i in set(frame.coded_species):
                # Get a randomized set of atoms of species i from the frame
                # So that it is not always the lowest-indexed atoms chosen
                atoms_of_specie = frame.indices_of_specie(species_i)
                n_spec = len(atoms_of_specie)
                # Determine how many to add based on user defined cutoffs
                n_to_add = min(
                    n_spec,
                    max_from_frame_by_species.get(species_i, inf),
                    max_atoms_from_frame,
                )

                train_atoms += random.sample(atoms_of_specie, n_spec)
                n_atoms_added += n_spec

            self.model.update_db(struc=frame, custom_range=train_atoms)
            if n_atoms_added >= max_atoms_added:
                break

        if n_atoms_added:
            logger.info(
                f"Added {n_atoms_added} atoms to "
                "pretrain.\n"
                "Pre-run GP Statistics: "
                f"{json.dumps(self.model.training_statistics)} "
            )

        if (envs or n_atoms_added or frames) and (
            self.pre_train_max_iter or self.max_trains
        ):
            logger.debug(
                "Now commencing pre-run training of GP (which has "
                "non-empty training set)"
            )
            time0 = time.time()
            self.train_gp(max_iter=self.pre_train_max_iter)
            logger.debug(f"Done train_gp {time.time()-time0}")
        else:
            logger.debug(
                "Now commencing pre-run set up of GP (which has non-empty training set)"
            )
            time0 = time.time()
            self.gp.check_L_alpha()
            logger.debug(f"Done check_L_alpha {time.time()-time0}")

        if self.model_format and not self.mgp:
            self.gp.write_model(f"{self.output_name}_prerun", self.model_format)


class TrainFromAIMD(LearningProtocol):
    def __init__(self):
        pass

    def train_model(self):
        pass


class LearningProtocol:
    def __init__(
        self,
        gp: Union[GaussianProcess, MappedGaussianProcess],
        active_frames: List[Structure] = None,
        passive_frames: List[Structure] = None,
        passive_envs: List[Tuple[AtomicEnvironment, "np.array"]] = None,
        active_rel_var_tol: float = 4,
        active_abs_var_tol: float = 1,
        active_abs_error_tol: float = 0,
        active_error_tol_cutoff: float = inf,
        active_max_trains: int = np.inf,
        active_max_element_from_frame: dict = None,
        checkpoint_interval_train: int = 1,
        checkpoint_interval_atom: int = 100,
        predict_atoms_per_element: dict = None,
        max_atoms_from_frame: int = np.inf,
        min_atoms_added_per_train: int = 1,
        max_model_size: int = np.inf,
        passive_on_active_skips: int = -1,
        passive_train_max_iter: int = 50,
        passive_atoms_per_element: dict = None,
        active_skip: int = 1,
        shuffle_active_frames: bool = False,
        n_cpus: int = 1,
        validate_ratio: float = 0.0,
        calculate_energy: bool = False,
        output_name: str = "gp_from_aimd",
        print_as_xyz: bool = False,
        verbose: str = "INFO",
        written_model_format: str = "json",
    ):
        """
        Class which trains a GP off of an AIMD trajectory, and generates
        error statistics between the DFT and GP calls.
        All arguments are divided between 'passive' learning and 'active'
        learning. By default, when run is called, a 'passive' learning run
        is called which either adds all 'seed' environments to the model,
        or a randomized subset of atoms from the frames. If no arguments are
        specified, the very first frame of the active learning
        frames will be used.
        "Passive" learning will add data based on random selection of atoms
        from a given ab-initio frame.
        "Active" learning will add data to the dataset based on the
        performance of the GP itself: the force error and the GP's internal
        uncertainty estimate.
        There are a widevariety of options which can give you a finer
        control over the training process.
        :param active_frames: List of structures to evaluate / train GP on
        :param gp: Gaussian Process object
        :param active_rel_var_tol: Train if uncertainty is above this *
            noise variance hyperparameter
        :param active_abs_var_tol: Train if uncertainty is above this
        :param active_abs_error_tol: Add atom force error exceeds this
        :param active_error_tol_cutoff: Don't add atom if force error exceeds this
        :param validate_ratio: Fraction of frames used for validation
        :param active_skip: Skip through frames
        :param calculate_energy: Use local energy kernel or not
        :param output_name: Write output of training to this file
        :param print_as_xyz: If True, print the configurations in xyz format
        :param max_atoms_from_frame: Largest # of atoms added from one frame
        :param min_atoms_added_per_train: Only train when this many atoms have been
            added
        :param active_max_trains: Stop training GP after this many calls to train
        :param n_cpus: Number of CPUs to parallelize over for parallelization
                over atoms
        :param shuffle_active_frames: Randomize order of frames for better training
        :param verbose: same as logging level, "WARNING", "INFO", "DEBUG"
        :param passive_on_active_skips: Train model on every n frames before running
        :param passive_frames: Frames to train on before running
        :param passive_envs: Environments to train on before running
        :param passive_atoms_per_element: Max # of environments to add from
            each species in the seed pre-training steps
        :param active_max_element_from_frame: Max # of environments to add from
            each species in the training steps
        :param predict_atoms_per_element: Choose a random subset of N random
            atoms from each specified element to predict on. For instance,
            {"H":5} will only predict the forces and uncertainties
            associated with 5 Hydrogen atoms per frame. Elements not
            specified will be predicted as normal. This is useful for
            systems where you are most interested in a subset of elements.
            This will result in a faster but less exhaustive learning process.
        :param checkpoint_interval_train: How often to write model after
                        trainings
        :param checkpoint_interval_atom: How often to write model after atoms are
            added (since atoms may be added without training)
        :param written_model_format: Format to write GP model to
        """

        # GP Training and Execution parameters
        self.gp = gp
        # Check to see if GP is MGP for later flagging
        self.mgp = isinstance(gp, MappedGaussianProcess)

        self.rel_std_tolerance = active_rel_var_tol
        self.abs_std_tolerance = active_abs_var_tol
        self.abs_force_tolerance = active_abs_error_tol
        self.max_force_error = active_error_tol_cutoff
        self.max_trains = active_max_trains
        self.max_atoms_from_frame = max_atoms_from_frame
        self.min_atoms_per_train = min_atoms_added_per_train
        self.max_model_size = max_model_size

        # Set prediction function based on if forces or energies are
        # desired, and parallelization accordingly
        if not self.mgp:
            if calculate_energy:
                self.pred_func = predict_on_structure_par_en
            else:
                self.pred_func = predict_on_structure_par

        elif self.mgp:
            self.pred_func = predict_on_structure_mgp

        self.start_time = time.time()

        self.train_count = 0
        self.calculate_energy = calculate_energy
        self.n_cpus = n_cpus

        # Output parameters
        self.output = Output(
            output_name, verbose, print_as_xyz=print_as_xyz, always_flush=True
        )
        self.logger_name = self.output.basename + "log"
        self.train_checkpoint_interval = checkpoint_interval_train
        self.atom_checkpoint_interval = checkpoint_interval_atom

        self.model_format = written_model_format
        self.output_name = output_name

        # gpfa only function

        self.predict_atoms_per_element = predict_atoms_per_element

        # Set up parameters
        self.frames = active_frames
        if shuffle_active_frames:
            np.random.shuffle(active_frames)

        # Parameters for negotiating with the training active_frames
        self.skip = active_skip
        assert (
            isinstance(active_skip, int) and active_skip >= 1
        ), "Skip needs to be a  positive integer."
        self.validate_ratio = validate_ratio
        assert 0 <= validate_ratio <= 1, "validate_ratio needs to be [0,1]"

        # Set up for pretraining
        self.pre_train_max_iter = passive_train_max_iter
        self.pre_train_on_skips = passive_on_active_skips
        self.seed_envs = [] if passive_envs is None else passive_envs
        self.seed_frames = [] if passive_frames is None else passive_frames

        self.pre_train_env_per_species = (
            {} if passive_atoms_per_element is None else passive_atoms_per_element
        )
        self.train_env_per_species = (
            {}
            if active_max_element_from_frame is None
            else active_max_element_from_frame
        )

        # Convert to Coded Species
        if self.pre_train_env_per_species:
            pre_train_species = list(self.pre_train_env_per_species.keys())
            for key in pre_train_species:
                self.pre_train_env_per_species[
                    element_to_Z(key)
                ] = self.pre_train_env_per_species[key]

        # Defining variables to be used later
        self.curr_step = 0
        self.train_count = 0
        self.start_time = time.time()

    def get_next_env(self):
        self.curr_env_index += 1
        if self.curr_env_index < len(self.seed_envs):
            return self.seed_envs[self.curr_env_index]
        return None

    def get_next_passive_frame(self):
        self.curr_passive_frame_index += 1
        if self.curr_passive_frame_index < len(self.seed_frames):
            return self.seed_frames[self.curr_passive_frame_index]
        return None

    def preparation_for_passive_run(self):
        # Remove frames used as seed from later part of training
        if self.pre_train_on_skips > 0:
            self.seed_frames = []
            newframes = []
            for i in range(len(self.frames)):
                if (i % self.pre_train_on_skips) == 0:
                    self.seed_frames += [self.frames[i]]
                else:
                    newframes += [self.frames[i]]
            self.frames = newframes
        # If the GP is empty, use the first frame as a seed frame.
        elif len(self.gp.training_data) == 0 and len(self.seed_frames) == 0:
            self.seed_frames = [self.frames[0]]
            self.frames = self.frames[1:]

    def preparation_for_active_run(self):
        raise NotImplementedError("need to be implemented in child class")

    def get_next_active_frame(self):
        raise NotImplementedError("need to be implemented in child class")

    def decide_to_update_db(self):
        raise NotImplementedError("need to be implemented in child class")

    def decide_to_checkLalpha(self):
        raise NotImplementedError("need to be implemented in child class")

    def passive_run(self):
        """
        Various tasks to set up the AIMD training before commencing
        the run through the AIMD trajectory.
        1. Print the output.
        2. Pre-train the GP with the seed frames and
        environments. If no seed frames or environments and the GP has no
        training set, then seed with at least one atom from each
        """

        if self.mgp:
            raise NotImplementedError("Pre-running notyet configured for MGP")
        self.output.write_header(
            str(self.gp),
            dt=0,
            Nsteps=len(self.frames),
            structure=None,
            std_tolerance=(self.rel_std_tolerance, self.abs_std_tolerance),
            optional={
                "GP Statistics": json.dumps(self.gp.training_statistics),
                "GP Name": self.gp.name,
                "GP Write Name": self.output_name + "_model." + self.model_format,
            },
        )

        self.start_time = time.time()
        logger = logging.getLogger(self.logger_name)
        logger.debug("Now beginning pre-run activity.")

        # If seed environments were passed in, add them to the GP.

        self.preparation_for_passive_run()

        self.curr_env_index = -1
        curr_env = self.get_next_env()
        while curr_env is not None:
            self.gp.add_one_env(curr_env[0], curr_env[1], train=False)
            curr_env = self.get_next_env()

        # No training set ("blank slate" run) and no seeds specified:
        # Take one of each atom species in the first frame
        # so all atomic species are represented in the first step.
        # Otherwise use the seed frames passed in by user.

        self.passive_atom_count = 0
        self.curr_passive_frame_index = -1
        frame = self.get_next_passive_frame()
        while frame is not None:

            train_atoms = []
            for species_i in set(frame.coded_species):
                # Get a randomized set of atoms of species i from the frame
                # So that it is not always the lowest-indexed atoms chosen
                atoms_of_specie = frame.indices_of_specie(species_i)
                np.random.shuffle(atoms_of_specie)
                n_at = len(atoms_of_specie)
                # Determine how many to add based on user defined cutoffs
                n_to_add = min(
                    n_at,
                    self.pre_train_env_per_species.get(species_i, inf),
                    self.max_atoms_from_frame,
                )

                for atom in atoms_of_specie[:n_to_add]:
                    train_atoms.append(atom)
                    self.passive_atom_count += 1

            self.update_gp_and_print(
                frame=frame, train_atoms=train_atoms, uncertainties=[], train=False
            )
            frame = self.get_next_passive_frame()

        logger = logging.getLogger(self.logger_name)
        if self.passive_atom_count > 0:
            logger.info(
                f"Added {self.passive_atom_count} atoms to "
                "pretrain.\n"
                "Pre-run GP Statistics: "
                f"{json.dumps(self.gp.training_statistics)} "
            )

        if (self.seed_envs or self.passive_atom_count or self.seed_frames) and (
            self.pre_train_max_iter or self.max_trains
        ):
            logger.debug(
                "Now commencing pre-run training of GP (which has "
                "non-empty training set)"
            )
            time0 = time.time()
            self.train_gp(max_iter=self.pre_train_max_iter)
            logger.debug(f"Done train_gp {time.time()-time0}")
        else:
            logger.debug(
                "Now commencing pre-run set up of GP (which has non-empty training set)"
            )
            time0 = time.time()
            self.gp.check_L_alpha()
            logger.debug(f"Done check_L_alpha {time.time()-time0}")

        if self.model_format and not self.mgp:
            self.gp.write_model(f"{self.output_name}_prerun", self.model_format)

    def active_run(self):
        """
        Loop through frames and record the error between
        the GP predictions and the ground-truth forces. Train the GP and update
        the training set upon the triggering of the uncertainty or force error
        threshold.
        :return: None
        """

        # Perform pre-run, in which seed trames are used.
        logger = logging.getLogger(self.logger_name)
        logger.debug("Commencing run with pre-run...")
        if not self.mgp:
            if len(self.gp) == 0:
                logger.warning(
                    "You are attempting to train a model with no "
                    "data in your Gausian Process; it is "
                    "recommended that you begin with "
                    "a passive training process."
                )

        self.preparation_for_active_run()

        # Loop through trajectory.
        self.cur_atoms_added_train = 0  # Track atoms added for training
        cur_atoms_added_write = 0  # Track atoms added for writing
        cur_trains_done_write = 0  # Track training done for writing

        self.curr_active_frame_index = -1
        cur_frame = self.get_next_active_frame()
        while cur_frame is not None:

            frame_start_time = time.time()
            logger.info(f"=====NOW ON FRAME {self.curr_active_frame_index}=====")

            # If no predict_atoms_per_element was specified, predict_atoms
            # will be equal to every atom in the frame.
            predict_atoms = subset_of_frame_by_element(
                cur_frame, self.predict_atoms_per_element
            )

            # Atoms which are skipped will have NaN as their force / std values
            local_energies = None

            # Three different predictions: Either MGP, GP with energy,
            # or GP without
            if self.mgp:
                pred_forces, pred_stds, local_energies = self.pred_func(
                    structure=cur_frame,
                    mgp=self.gp,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                    energy=True,
                )
            elif self.calculate_energy:
                pred_forces, pred_stds, local_energies = self.pred_func(
                    structure=cur_frame,
                    gp=self.gp,
                    n_cpus=self.n_cpus,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                )
            else:
                pred_forces, pred_stds = self.pred_func(
                    structure=cur_frame,
                    gp=self.gp,
                    n_cpus=self.n_cpus,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                )

            # Get Error
            dft_forces = cur_frame.forces
            dft_energy = cur_frame.energy
            error = np.abs(pred_forces - dft_forces)

            # Create dummy frame with the predicted forces written
            dummy_frame = deepcopy(cur_frame)
            dummy_frame.forces = pred_forces
            dummy_frame.stds = pred_stds

            self.output.write_gp_dft_comparison(
                curr_step=self.curr_active_frame_index,
                frame=dummy_frame,
                start_time=time.time(),
                dft_forces=dft_forces,
                dft_energy=dft_energy,
                error=error,
                local_energies=local_energies,
                KE=0,
            )

            logger.debug(
                f"Single frame calculation time {time.time()-frame_start_time}"
            )

            if self.decide_to_update_db():

                # Noise hyperparameter & relative std tolerance is not for mgp.
                if self.mgp:
                    noise = 0
                else:
                    noise = Parameters.get_noise(
                        self.gp.hyps_mask, self.gp.hyps, constraint=False
                    )

                std_in_bound, std_train_atoms = is_std_in_bound_per_species(
                    rel_std_tolerance=self.rel_std_tolerance,
                    abs_std_tolerance=self.abs_std_tolerance,
                    noise=noise,
                    structure=dummy_frame,
                    max_atoms_added=self.max_atoms_from_frame,
                    max_by_species=self.train_env_per_species,
                )

                # Get max force error atoms
                force_in_bound, force_train_atoms = is_force_in_bound_per_species(
                    abs_force_tolerance=self.abs_force_tolerance,
                    predicted_forces=pred_forces,
                    label_forces=dft_forces,
                    structure=dummy_frame,
                    max_atoms_added=self.max_atoms_from_frame,
                    max_by_species=self.train_env_per_species,
                    max_force_error=self.max_force_error,
                )

                if not std_in_bound or not force_in_bound:

                    # -1 is returned from the is_in_bound methods,
                    # so filter that out and the use sets to remove repeats
                    train_atoms = list(
                        set(std_train_atoms).union(force_train_atoms) - {-1}
                    )

                    # Compute mae and write to output;
                    # Add max uncertainty atoms to training set
                    self.update_gp_and_print(
                        cur_frame,
                        train_atoms=train_atoms,
                        uncertainties=pred_stds[train_atoms],
                        train=False,
                    )
                    self.cur_atoms_added_train += len(train_atoms)
                    cur_atoms_added_write += len(train_atoms)
                    # Re-train if number of sampled atoms is high enough

                    if self.decide_to_train():
                        self.train_gp()
                        cur_trains_done_write += 1
                        self.cur_atoms_added_train = 0
                    else:
                        self.gp.update_L_alpha()
                        # self.cur_atoms_added_train = 0

                    # Loop to decide of a model should be written this
                    # iteration
                    will_write = False

                    if (
                        self.train_checkpoint_interval
                        and cur_trains_done_write
                        and self.train_checkpoint_interval <= cur_trains_done_write
                    ):
                        will_write = True
                        cur_trains_done_write = 0

                    if (
                        self.atom_checkpoint_interval
                        and cur_atoms_added_write
                        and self.atom_checkpoint_interval <= cur_atoms_added_write
                    ):
                        will_write = True
                        cur_atoms_added_write = 0

                    if self.model_format and will_write:
                        self.gp.write_model(
                            f"{self.output_name}_checkpt", self.model_format
                        )

                if self.decide_to_checkLalpha():
                    self.gp.check_L_alpha()

            cur_frame = self.get_next_active_frame()

        self.output.conclude_run()

        if self.model_format and not self.mgp:
            self.gp.write_model(f"{self.output_name}_model", self.model_format)

    def update_gp_and_print(
        self,
        frame: Structure,
        train_atoms: List[int],
        uncertainties: List[int] = None,
        train: bool = True,
    ):
        """
        Update the internal GP model training set with a list of training
        atoms indexing atoms within the frame. If train is True, re-train
        the GP by optimizing hyperparameters.
        :param frame: Structure to train on
        :param train_atoms: Index atoms to train on
        :param uncertainties: Uncertainties to print, pass in [] to silence
        :param train: Train or not
        :return: None
        """

        # Group added atoms by species for easier output
        added_species = [Z_to_element(frame.coded_species[at]) for at in train_atoms]
        added_atoms = {spec: [] for spec in set(added_species)}

        for atom, spec in zip(train_atoms, added_species):
            added_atoms[spec].append(atom)

        logger = logging.getLogger(self.logger_name)
        logger.info(
            "Adding atom(s) "
            f"{json.dumps(added_atoms,cls=NumpyEncoder)}"
            " to the training set."
        )

        if uncertainties is None or len(uncertainties) != 0:
            uncertainties = frame.stds[train_atoms]

        if len(uncertainties) != 0:
            logger.info(f"Uncertainties: {uncertainties}.")

        # update gp model; handling differently if it's an MGP
        if not self.mgp:
            self.gp.update_db(frame, frame.forces, custom_range=train_atoms)

            if train:
                self.train_gp()

        else:
            raise NotImplementedError

    def train_gp(self, max_iter: int = None):
        """
        Train the Gaussian process and write the results to the output file.
        :param max_iter: Maximum iterations associated with this training run,
            overriding the Gaussian Process's internally set maxiter.
        :type max_iter: int
        """

        logger = logging.getLogger(self.logger_name)
        logger.debug("Train GP")

        logger_train = self.output.basename + "hyps"

        # TODO: Improve flexibility in GP training to make this next step
        # unnecessary, so maxiter can be passed as an argument

        # Don't train if maxiter == 0
        if max_iter == 0:
            self.gp.check_L_alpha()
        elif max_iter is not None:
            temp_maxiter = self.gp.maxiter
            self.gp.maxiter = max_iter
            self.gp.train(logger_name=logger_train)
            self.gp.maxiter = temp_maxiter
        else:
            self.gp.train(logger_name=logger_train)

        hyps, labels = Parameters.get_hyps(
            self.gp.hyps_mask, self.gp.hyps, constraint=False, label=True
        )
        if labels is None:
            labels = self.gp.hyp_labels
        self.output.write_hyps(
            labels,
            hyps,
            self.start_time,
            self.gp.likelihood,
            self.gp.likelihood_gradient,
            hyps_mask=self.gp.hyps_mask,
        )
        self.train_count += 1


def parse_trajectory_trainer_output(
    file: str, return_gp_data: bool = False
) -> Union[List[dict], Tuple[List[dict], dict]]:
    """
    Reads output of a TrajectoryTrainer run by frame. return_gp_data returns
    data about GP model growth useful for visualizing progress of model
    training.
    :param file: filename of output
    :param return_gp_data: flag for returning extra GP data
    :return: List of dictionaries with keys 'species', 'positions',
        'gp_forces', 'dft_forces', 'gp_stds', 'added_atoms', and
        'maes_by_species', optionally, gp_data dictionary
    """

    with open(file, "r") as f:
        lines = f.readlines()
        num_lines = len(lines)

    # Get indexes where frames begin, and include the index of the final line
    frame_indexes = [i for i in range(num_lines) if "-Frame:" in lines[i]] + [num_lines]

    frames = []

    # Central parsing loop
    for n in range(len(frame_indexes) - 1):
        # Start at +2 to skip frame marker and header of table of data
        # Set up values for current frame which will be populated

        frame_atoms = []
        frame_positions = []
        gp_forces = []
        dft_forces = []
        stds = []
        added_atoms = {}
        frame_species_maes = {}

        # i loops through individual atom's info
        for i in range(frame_indexes[n] + 2, frame_indexes[n + 1]):

            # Lines with data will be long; stop when at end of atom data
            if len(lines[i]) > 10:
                split = lines[i].split()

                frame_atoms.append(split[0])

                frame_positions.append(
                    [float(split[1]), float(split[2]), float(split[3])]
                )
                gp_forces.append([float(split[4]), float(split[5]), float(split[6])])
                stds.append([float(split[7]), float(split[8]), float(split[9])])

                dft_forces.append(
                    [float(split[10]), float(split[11]), float(split[12])]
                )

            # Terminate at blank line between results
            else:
                break
        # Loop through information in frame after Data
        for i in range(
            frame_indexes[n] + len(frame_positions) + 2, frame_indexes[n + 1]
        ):

            if "Adding atom(s)" in lines[i]:
                # Splitting to target the 'added atoms' substring
                split_line = lines[i][15:-21]
                added_atoms = json.loads(split_line.strip())

            if "type " in lines[i]:
                cur_line = lines[i].split()
                frame_species_maes[cur_line[1]] = float(cur_line[3])

        cur_frame_stats = {
            "species": frame_atoms,
            "positions": frame_positions,
            "gp_forces": gp_forces,
            "dft_forces": dft_forces,
            "gp_stds": stds,
            "added_atoms": added_atoms,
            "maes_by_species": frame_species_maes,
        }

        frames.append(cur_frame_stats)

    if not return_gp_data:
        return frames

    # Compute information about GP training
    # to study GP growth and performance over trajectory

    gp_stats_line = [
        line for line in lines[:30] if "GP Statistics" in line and "Pre-run" not in line
    ][0][15:].strip()

    initial_gp_statistics = json.loads(gp_stats_line)

    # Get pre_run statistics (if pre-run was done):
    pre_run_gp_statistics = None
    pre_run_gp_stats_line = [line for line in lines if "Pre-run GP" in line]
    if pre_run_gp_stats_line:
        pre_run_gp_statistics = json.loads(pre_run_gp_stats_line[0][22:].strip())

    # Compute cumulative GP size
    cumulative_gp_size = [int(initial_gp_statistics["N"])]

    if pre_run_gp_stats_line:
        cumulative_gp_size.append(int(pre_run_gp_statistics["N"]))

    running_total = cumulative_gp_size[-1]

    for frame in frames:

        added_atom_dict = frame["added_atoms"]
        for val in added_atom_dict.values():
            running_total += len(val)
        cumulative_gp_size.append(running_total)

    # Compute MAEs for each element over time
    all_species = set()
    for frame in frames:
        all_species = all_species.union(set(frame["species"]))

    all_species = list(all_species)
    mae_by_elt = {elt: [] for elt in all_species}

    for frame in frames:
        for elt in all_species:
            cur_mae = frame["maes_by_species"].get(elt, np.nan)
            mae_by_elt[elt].append(cur_mae)

    gp_data = {
        "init_stats": initial_gp_statistics,
        "pre_train_stats": pre_run_gp_statistics,
        "cumulative_gp_size": cumulative_gp_size,
        "mae_by_elt": mae_by_elt,
    }

    return frames, gp_data



import numpy as np
from flare.kernels.kernels import (
    force_helper,
    force_energy_helper,
    grad_helper,
    three_body_fe_perm,
    three_body_ee_perm,
    three_body_se_perm,
    three_body_ff_perm,
    three_body_sf_perm,
    three_body_ss_perm,
    three_body_grad_perm,
    grad_constants,
)
from numba import njit
from flare.env import AtomicEnvironment
from typing import Callable
import flare.kernels.cutoffs as cf
from math import exp


class ThreeBodyKernel:
    def __init__(
        self,
        hyperparameters: "ndarray",
        cutoff: float,
        cutoff_func: Callable = cf.quadratic_cutoff,
    ):
        self.hyperparameters = hyperparameters
        self.signal_variance = hyperparameters[0]
        self.length_scale = hyperparameters[1]
        self.cutoff = cutoff
        self.cutoff_func = cutoff_func

    def energy_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return energy_energy(*args)

    def force_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_energy(*args)

    def stress_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_energy(*args)

    def force_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_force(*args)

    def stress_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_force(*args)

    def stress_stress(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return stress_stress(*args)

    def force_force_gradient(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return force_force_gradient(*args)

    def efs_energy(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return efs_energy(*args)

    def efs_force(self, env1: AtomicEnvironment, env2: AtomicEnvironment):
        args = self.get_args(env1, env2)
        return efs_force(*args)

    def efs_self(self, env1: AtomicEnvironment):
        return efs_self(
            env1.bond_array_3,
            env1.ctype,
            env1.etypes,
            env1.cross_bond_inds,
            env1.cross_bond_dists,
            env1.triplet_counts,
            self.signal_variance,
            self.length_scale,
            self.cutoff,
            self.cutoff_func,
        )

    def get_args(self, env1, env2):
        return (
            env1.bond_array_3,
            env1.ctype,
            env1.etypes,
            env2.bond_array_3,
            env2.ctype,
            env2.etypes,
            env1.cross_bond_inds,
            env2.cross_bond_inds,
            env1.cross_bond_dists,
            env2.cross_bond_dists,
            env1.triplet_counts,
            env2.triplet_counts,
            self.signal_variance,
            self.length_scale,
            self.cutoff,
            self.cutoff_func,
        )


@njit
def energy_energy(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between two local energies accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 3-body local energy kernel.
    """

    kern = 0

    sig2 = sig * sig
    ls2 = 1 / (2 * ls * ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1 * fi2 * fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes2[p]

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    ej2 = etypes2[ind2]

                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1 * fj2 * fj3

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    kern += three_body_ee_perm(
                        r11,
                        r12,
                        r13,
                        r21,
                        r22,
                        r23,
                        r31,
                        r32,
                        r33,
                        c1,
                        c2,
                        ei1,
                        ei2,
                        ej1,
                        ej2,
                        fi,
                        fj,
                        ls2,
                        sig2,
                    )

    return kern / 9


@njit
def force_energy(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between a force component and a local
    energy accelerated with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 3-body force/energy kernel.
    """
    kern = np.zeros(3)

    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes2[p]

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    ej2 = etypes2[ind2]
                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1 * fj2 * fj3

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                        kern[d1] += three_body_fe_perm(
                            r11,
                            r12,
                            r13,
                            r21,
                            r22,
                            r23,
                            r31,
                            r32,
                            r33,
                            c1,
                            c2,
                            ci1,
                            ci2,
                            ei1,
                            ei2,
                            ej1,
                            ej2,
                            fi,
                            fj,
                            fdi,
                            ls1,
                            ls2,
                            sig2,
                        )

    return kern / 3


@njit
def stress_energy(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between a force component and a local
    energy accelerated with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        float:
            Value of the 3-body force/energy kernel.
    """
    kern = np.zeros(6)

    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1 * fi2 * fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes2[p]

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    ej2 = etypes2[ind2]
                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1 * fj2 * fj3

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    stress_count = 0
                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fdi_p1 = fdi1 * fi2 * fi3
                        fdi_p2 = fi1 * fdi2 * fi3
                        fdi = fdi_p1 + fdi_p2

                        for d2 in range(d1, 3):
                            coord1 = bond_array_1[m, d2 + 1] * ri1
                            coord2 = bond_array_1[ind1, d2 + 1] * ri2

                            kern[stress_count] += three_body_se_perm(
                                r11,
                                r12,
                                r13,
                                r21,
                                r22,
                                r23,
                                r31,
                                r32,
                                r33,
                                c1,
                                c2,
                                ci1,
                                ci2,
                                ei1,
                                ei2,
                                ej1,
                                ej2,
                                fi,
                                fj,
                                fdi,
                                ls1,
                                ls2,
                                sig2,
                                coord1,
                                coord2,
                                fdi_p1,
                                fdi_p2,
                            )

                            stress_count += 1

    return kern / 6


@njit
def force_force(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 3-body kernel.
    """
    kern = np.zeros((3, 3))

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    # first loop over the first 3-body environment
    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        # second loop over the first 3-body environment
        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            # first loop over the second 3-body environment
            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                ej1 = etypes2[p]

                # second loop over the second 3-body environment
                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + 1 + q]
                    rj2 = bond_array_2[ind2, 0]
                    rj3 = cross_bond_dists_2[p, p + 1 + q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    ej2 = etypes2[ind2]

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                        for d2 in range(3):
                            cj1 = bond_array_2[p, d2 + 1]
                            fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                            cj2 = bond_array_2[ind2, d2 + 1]
                            fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                            fj = fj1 * fj2 * fj3
                            fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                            kern[d1, d2] += three_body_ff_perm(
                                r11,
                                r12,
                                r13,
                                r21,
                                r22,
                                r23,
                                r31,
                                r32,
                                r33,
                                c1,
                                c2,
                                ci1,
                                ci2,
                                cj1,
                                cj2,
                                ei1,
                                ei2,
                                ej1,
                                ej2,
                                fi,
                                fj,
                                fdi,
                                fdj,
                                ls1,
                                ls2,
                                ls3,
                                sig2,
                            )

    return kern


@njit
def stress_force(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 3-body kernel.
    """
    kern = np.zeros((6, 3))

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    # first loop over the first 3-body environment
    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        # second loop over the first 3-body environment
        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            # first loop over the second 3-body environment
            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                ej1 = etypes2[p]

                # second loop over the second 3-body environment
                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + 1 + q]
                    rj2 = bond_array_2[ind2, 0]
                    rj3 = cross_bond_dists_2[p, p + 1 + q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    ej2 = etypes2[ind2]

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    stress_count = 0
                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi_p1 = fdi1 * fi2 * fi3
                        fdi_p2 = fi1 * fdi2 * fi3
                        fdi = fdi_p1 + fdi_p2

                        for d2 in range(d1, 3):
                            coord1 = bond_array_1[m, d2 + 1] * ri1
                            coord2 = bond_array_1[ind1, d2 + 1] * ri2

                            for d3 in range(3):
                                cj1 = bond_array_2[p, d3 + 1]
                                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                                cj2 = bond_array_2[ind2, d3 + 1]
                                fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                                fj = fj1 * fj2 * fj3
                                fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                                kern[stress_count, d3] += three_body_sf_perm(
                                    r11,
                                    r12,
                                    r13,
                                    r21,
                                    r22,
                                    r23,
                                    r31,
                                    r32,
                                    r33,
                                    c1,
                                    c2,
                                    ci1,
                                    ci2,
                                    cj1,
                                    cj2,
                                    ei1,
                                    ei2,
                                    ej1,
                                    ej2,
                                    fi,
                                    fj,
                                    fdi,
                                    fdj,
                                    ls1,
                                    ls2,
                                    ls3,
                                    sig2,
                                    coord1,
                                    coord2,
                                    fdi_p1,
                                    fdi_p2,
                                )
                            stress_count += 1

    return kern / 2


@njit
def stress_stress(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the 3-body kernel.
    """
    kern = np.zeros((6, 6))

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    # first loop over the first 3-body environment
    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        # second loop over the first 3-body environment
        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            # first loop over the second 3-body environment
            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                ej1 = etypes2[p]

                # second loop over the second 3-body environment
                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + 1 + q]
                    rj2 = bond_array_2[ind2, 0]
                    rj3 = cross_bond_dists_2[p, p + 1 + q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    ej2 = etypes2[ind2]

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    stress_count_1 = 0
                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi_p1 = fdi1 * fi2 * fi3
                        fdi_p2 = fi1 * fdi2 * fi3
                        fdi = fdi_p1 + fdi_p2

                        for d2 in range(d1, 3):
                            coord1 = bond_array_1[m, d2 + 1] * ri1
                            coord2 = bond_array_1[ind1, d2 + 1] * ri2

                            stress_count_2 = 0
                            for d3 in range(3):
                                cj1 = bond_array_2[p, d3 + 1]
                                fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                                cj2 = bond_array_2[ind2, d3 + 1]
                                fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                                fj = fj1 * fj2 * fj3
                                fdj_p1 = fdj1 * fj2 * fj3
                                fdj_p2 = fj1 * fdj2 * fj3
                                fdj = fdj_p1 + fdj_p2

                                for d4 in range(d3, 3):
                                    coord3 = bond_array_2[p, d4 + 1] * rj1
                                    coord4 = bond_array_2[ind2, d4 + 1] * rj2

                                    kern[
                                        stress_count_1, stress_count_2
                                    ] += three_body_ss_perm(
                                        r11,
                                        r12,
                                        r13,
                                        r21,
                                        r22,
                                        r23,
                                        r31,
                                        r32,
                                        r33,
                                        c1,
                                        c2,
                                        ci1,
                                        ci2,
                                        cj1,
                                        cj2,
                                        ei1,
                                        ei2,
                                        ej1,
                                        ej2,
                                        fi,
                                        fj,
                                        fdi,
                                        fdj,
                                        ls1,
                                        ls2,
                                        ls3,
                                        sig2,
                                        coord1,
                                        coord2,
                                        coord3,
                                        coord4,
                                        fdi_p1,
                                        fdi_p2,
                                        fdj_p1,
                                        fdj_p2,
                                    )
                                    stress_count_2 += 1
                            stress_count_1 += 1

    return kern / 4


@njit
def force_force_gradient(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):
    """3-body multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        bond_array_1 (np.ndarray): 3-body bond array of the first local
            environment.
        c1 (int): Species of the central atom of the first local environment.
        etypes1 (np.ndarray): Species of atoms in the first local
            environment.
        bond_array_2 (np.ndarray): 3-body bond array of the second local
            environment.
        c2 (int): Species of the central atom of the second local environment.
        etypes2 (np.ndarray): Species of atoms in the second local
            environment.
        cross_bond_inds_1 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the first local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_inds_2 (np.ndarray): Two dimensional array whose row m
            contains the indices of atoms n > m in the second local
            environment that are within a distance r_cut of both atom n and
            the central atom.
        cross_bond_dists_1 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the first
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        cross_bond_dists_2 (np.ndarray): Two dimensional array whose row m
            contains the distances from atom m of atoms n > m in the second
            local environment that are within a distance r_cut of both atom
            n and the central atom.
        triplets_1 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the first local environment that are
            within a distance r_cut of atom m.
        triplets_2 (np.ndarray): One dimensional array of integers whose entry
            m is the number of atoms in the second local environment that are
            within a distance r_cut of atom m.
        sig (float): 3-body signal variance hyperparameter.
        ls (float): 3-body length scale hyperparameter.
        r_cut (float): 3-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Returns:
        (float, float):
            Value of the 3-body kernel and its gradient with respect to the
            hyperparameters.
    """
    kernel_matrix = np.zeros((3, 3))
    kernel_grad = np.zeros((2, 3, 3))

    # pre-compute constants that appear in the inner loop
    sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6 = grad_constants(sig, ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri3 = cross_bond_dists_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            ei2 = etypes1[ind1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                ej1 = etypes2[p]

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    ej2 = etypes2[ind2]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3
                        fi = fi1 * fi2 * fi3

                        for d2 in range(3):
                            cj1 = bond_array_2[p, d2 + 1]
                            fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                            cj2 = bond_array_2[ind2, d2 + 1]
                            fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                            fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3
                            fj = fj1 * fj2 * fj3

                            kern_term, sig_term, ls_term = three_body_grad_perm(
                                r11,
                                r12,
                                r13,
                                r21,
                                r22,
                                r23,
                                r31,
                                r32,
                                r33,
                                c1,
                                c2,
                                ci1,
                                ci2,
                                cj1,
                                cj2,
                                ei1,
                                ei2,
                                ej1,
                                ej2,
                                fi,
                                fj,
                                fdi,
                                fdj,
                                ls1,
                                ls2,
                                ls3,
                                ls4,
                                ls5,
                                ls6,
                                sig2,
                                sig3,
                            )

                            kernel_matrix[d1, d2] += kern_term
                            kernel_grad[0, d1, d2] += sig_term
                            kernel_grad[1, d1, d2] += ls_term

    return kernel_matrix, kernel_grad


@njit
def efs_energy(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):

    energy_kernel = 0
    force_kernels = np.zeros(3)
    stress_kernels = np.zeros(6)

    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1 * fi2 * fi3

            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes2[p]

                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + q + 1]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    ej2 = etypes2[ind2]
                    rj3 = cross_bond_dists_2[p, p + q + 1]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1 * fj2 * fj3

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    energy_kernel += (
                        three_body_ee_perm(
                            r11,
                            r12,
                            r13,
                            r21,
                            r22,
                            r23,
                            r31,
                            r32,
                            r33,
                            c1,
                            c2,
                            ei1,
                            ei2,
                            ej1,
                            ej2,
                            fi,
                            fj,
                            ls1,
                            sig2,
                        )
                        / 9
                    )

                    stress_count = 0
                    for d1 in range(3):
                        ci1 = bond_array_1[m, d1 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d1 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi_p1 = fdi1 * fi2 * fi3
                        fdi_p2 = fi1 * fdi2 * fi3
                        fdi = fdi_p1 + fdi_p2

                        force_kernels[d1] += (
                            three_body_fe_perm(
                                r11,
                                r12,
                                r13,
                                r21,
                                r22,
                                r23,
                                r31,
                                r32,
                                r33,
                                c1,
                                c2,
                                ci1,
                                ci2,
                                ei1,
                                ei2,
                                ej1,
                                ej2,
                                fi,
                                fj,
                                fdi,
                                ls1,
                                ls2,
                                sig2,
                            )
                            / 3
                        )

                        for d2 in range(d1, 3):
                            coord1 = bond_array_1[m, d2 + 1] * ri1
                            coord2 = bond_array_1[ind1, d2 + 1] * ri2

                            stress_kernels[stress_count] += (
                                three_body_se_perm(
                                    r11,
                                    r12,
                                    r13,
                                    r21,
                                    r22,
                                    r23,
                                    r31,
                                    r32,
                                    r33,
                                    c1,
                                    c2,
                                    ci1,
                                    ci2,
                                    ei1,
                                    ei2,
                                    ej1,
                                    ej2,
                                    fi,
                                    fj,
                                    fdi,
                                    ls1,
                                    ls2,
                                    sig2,
                                    coord1,
                                    coord2,
                                    fdi_p1,
                                    fdi_p2,
                                )
                                / 6
                            )

                            stress_count += 1

    return energy_kernel, force_kernels, stress_kernels


@njit
def efs_force(
    bond_array_1,
    c1,
    etypes1,
    bond_array_2,
    c2,
    etypes2,
    cross_bond_inds_1,
    cross_bond_inds_2,
    cross_bond_dists_1,
    cross_bond_dists_2,
    triplets_1,
    triplets_2,
    sig,
    ls,
    r_cut,
    cutoff_func,
):

    energy_kernels = np.zeros(3)
    force_kernels = np.zeros((3, 3))
    stress_kernels = np.zeros((6, 3))

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    # first loop over the first 3-body environment
    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        # second loop over the first 3-body environment
        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1 * fi2 * fi3

            # first loop over the second 3-body environment
            for p in range(bond_array_2.shape[0]):
                rj1 = bond_array_2[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes2[p]

                # second loop over the second 3-body environment
                for q in range(triplets_2[p]):
                    ind2 = cross_bond_inds_2[p, p + 1 + q]
                    rj2 = bond_array_2[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    rj3 = cross_bond_dists_2[p, p + 1 + q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    ej2 = etypes2[ind2]

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    for d3 in range(3):
                        cj1 = bond_array_2[p, d3 + 1]
                        fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                        cj2 = bond_array_2[ind2, d3 + 1]
                        fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                        fj = fj1 * fj2 * fj3
                        fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                        energy_kernels[d3] += (
                            three_body_fe_perm(
                                r11,
                                r21,
                                r31,
                                r12,
                                r22,
                                r32,
                                r13,
                                r23,
                                r33,
                                c2,
                                c1,
                                -cj1,
                                -cj2,
                                ej1,
                                ej2,
                                ei1,
                                ei2,
                                fj,
                                fi,
                                fdj,
                                ls1,
                                ls2,
                                sig2,
                            )
                            / 3
                        )

                        stress_count = 0
                        for d1 in range(3):
                            ci1 = bond_array_1[m, d1 + 1]
                            fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                            ci2 = bond_array_1[ind1, d1 + 1]
                            fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                            fi = fi1 * fi2 * fi3
                            fdi_p1 = fdi1 * fi2 * fi3
                            fdi_p2 = fi1 * fdi2 * fi3
                            fdi = fdi_p1 + fdi_p2

                            force_kernels[d1, d3] += three_body_ff_perm(
                                r11,
                                r12,
                                r13,
                                r21,
                                r22,
                                r23,
                                r31,
                                r32,
                                r33,
                                c1,
                                c2,
                                ci1,
                                ci2,
                                cj1,
                                cj2,
                                ei1,
                                ei2,
                                ej1,
                                ej2,
                                fi,
                                fj,
                                fdi,
                                fdj,
                                ls1,
                                ls2,
                                ls3,
                                sig2,
                            )

                            for d2 in range(d1, 3):
                                coord1 = bond_array_1[m, d2 + 1] * ri1
                                coord2 = bond_array_1[ind1, d2 + 1] * ri2

                                stress_kernels[stress_count, d3] += (
                                    three_body_sf_perm(
                                        r11,
                                        r12,
                                        r13,
                                        r21,
                                        r22,
                                        r23,
                                        r31,
                                        r32,
                                        r33,
                                        c1,
                                        c2,
                                        ci1,
                                        ci2,
                                        cj1,
                                        cj2,
                                        ei1,
                                        ei2,
                                        ej1,
                                        ej2,
                                        fi,
                                        fj,
                                        fdi,
                                        fdj,
                                        ls1,
                                        ls2,
                                        ls3,
                                        sig2,
                                        coord1,
                                        coord2,
                                        fdi_p1,
                                        fdi_p2,
                                    )
                                    / 2
                                )
                                stress_count += 1

    return energy_kernels, force_kernels, stress_kernels


@njit
def efs_self(
    bond_array_1,
    c1,
    etypes1,
    cross_bond_inds_1,
    cross_bond_dists_1,
    triplets_1,
    sig,
    ls,
    r_cut,
    cutoff_func,
):

    energy_kernel = 0
    force_kernels = np.zeros(3)
    stress_kernels = np.zeros(6)

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ri2 = bond_array_1[ind1, 0]
            fi2, _ = cutoff_func(r_cut, ri2, 0)
            ei2 = etypes1[ind1]

            ri3 = cross_bond_dists_1[m, m + n + 1]
            fi3, _ = cutoff_func(r_cut, ri3, 0)
            fi = fi1 * fi2 * fi3

            for p in range(bond_array_1.shape[0]):
                rj1 = bond_array_1[p, 0]
                fj1, _ = cutoff_func(r_cut, rj1, 0)
                ej1 = etypes1[p]

                for q in range(triplets_1[p]):
                    ind2 = cross_bond_inds_1[p, p + 1 + q]
                    rj2 = bond_array_1[ind2, 0]
                    fj2, _ = cutoff_func(r_cut, rj2, 0)
                    rj3 = cross_bond_dists_1[p, p + 1 + q]
                    fj3, _ = cutoff_func(r_cut, rj3, 0)
                    fj = fj1 * fj2 * fj3
                    ej2 = etypes1[ind2]

                    r11 = ri1 - rj1
                    r12 = ri1 - rj2
                    r13 = ri1 - rj3
                    r21 = ri2 - rj1
                    r22 = ri2 - rj2
                    r23 = ri2 - rj3
                    r31 = ri3 - rj1
                    r32 = ri3 - rj2
                    r33 = ri3 - rj3

                    energy_kernel += (
                        three_body_ee_perm(
                            r11,
                            r12,
                            r13,
                            r21,
                            r22,
                            r23,
                            r31,
                            r32,
                            r33,
                            c1,
                            c1,
                            ei1,
                            ei2,
                            ej1,
                            ej2,
                            fi,
                            fj,
                            ls1,
                            sig2,
                        )
                        / 9
                    )

                    stress_count = 0
                    for d3 in range(3):
                        cj1 = bond_array_1[p, d3 + 1]
                        fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)
                        cj2 = bond_array_1[ind2, d3 + 1]
                        fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                        fdj_p1 = fdj1 * fj2 * fj3
                        fdj_p2 = fj1 * fdj2 * fj3
                        fdj = fdj_p1 + fdj_p2

                        ci1 = bond_array_1[m, d3 + 1]
                        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
                        ci2 = bond_array_1[ind1, d3 + 1]
                        fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                        fi = fi1 * fi2 * fi3
                        fdi_p1 = fdi1 * fi2 * fi3
                        fdi_p2 = fi1 * fdi2 * fi3
                        fdi = fdi_p1 + fdi_p2

                        force_kernels[d3] += three_body_ff_perm(
                            r11,
                            r12,
                            r13,
                            r21,
                            r22,
                            r23,
                            r31,
                            r32,
                            r33,
                            c1,
                            c1,
                            ci1,
                            ci2,
                            cj1,
                            cj2,
                            ei1,
                            ei2,
                            ej1,
                            ej2,
                            fi,
                            fj,
                            fdi,
                            fdj,
                            ls1,
                            ls2,
                            ls3,
                            sig2,
                        )

                        for d2 in range(d3, 3):
                            coord1 = bond_array_1[m, d2 + 1] * ri1
                            coord2 = bond_array_1[ind1, d2 + 1] * ri2
                            coord3 = bond_array_1[p, d2 + 1] * rj1
                            coord4 = bond_array_1[ind2, d2 + 1] * rj2

                            stress_kernels[stress_count] += (
                                three_body_ss_perm(
                                    r11,
                                    r12,
                                    r13,
                                    r21,
                                    r22,
                                    r23,
                                    r31,
                                    r32,
                                    r33,
                                    c1,
                                    c1,
                                    ci1,
                                    ci2,
                                    cj1,
                                    cj2,
                                    ei1,
                                    ei2,
                                    ej1,
                                    ej2,
                                    fi,
                                    fj,
                                    fdi,
                                    fdj,
                                    ls1,
                                    ls2,
                                    ls3,
                                    sig2,
                                    coord1,
                                    coord2,
                                    coord3,
                                    coord4,
                                    fdi_p1,
                                    fdi_p2,
                                    fdj_p1,
                                    fdj_p2,
                                )
                                / 4
                            )
                            stress_count += 1

    return energy_kernel, force_kernels, stress_kernels



"""
:class:`ASE_OTF` is the on-the-fly training module for ASE, WITHOUT molecular dynamics engine. 
It needs to be used adjointly with ASE MD engine. 
"""
import os
import json
import sys
import inspect
import pickle
from time import time
from copy import deepcopy
import logging

import numpy as np
from flare.ase.npt import NPT_mod
from ase.md.nvtberendsen import NVTBerendsen
from ase.md.nptberendsen import NPTBerendsen
from ase.md.verlet import VelocityVerlet
from ase.md.langevin import Langevin
from flare.ase.nosehoover import NoseHoover
from ase import units
from ase.io import read, write

import flare
from flare.otf import OTF
from flare.ase.atoms import FLARE_Atoms
from flare.ase.calculator import FLARE_Calculator
import flare.ase.dft as dft_source


class ASE_OTF(OTF):

    """
    On-the-fly training module using ASE MD engine, a subclass of OTF.

    Args:
        atoms (ASE Atoms): the ASE Atoms object for the on-the-fly MD run.
        calculator: ASE calculator. Must have "get_uncertainties" method
          implemented.
        timestep: the timestep in MD. Please use ASE units, e.g. if the
            timestep is 1 fs, then set `timestep = 1 * units.fs`
        number_of_steps (int): the total number of steps for MD.
        dft_calc (ASE Calculator): any ASE calculator is supported,
            e.g. Espresso, VASP etc.
        md_engine (str): the name of MD thermostat, only `VelocityVerlet`,
            `NVTBerendsen`, `NPTBerendsen`, `NPT` and `Langevin`, `NoseHoover`
            are supported.
        md_kwargs (dict): Specify the args for MD as a dictionary, the args are
            as required by the ASE MD modules consistent with the `md_engine`.
        trajectory (ASE Trajectory): default `None`, not recommended,
            currently in experiment.

    The following arguments are for on-the-fly training, the user can also
    refer to :class:`flare.otf.OTF`

    Args:
        prev_pos_init ([type], optional): Previous positions. Defaults
            to None.
        rescale_steps (List[int], optional): List of frames for which the
            velocities of the atoms are rescaled. Defaults to [].
        rescale_temps (List[int], optional): List of rescaled temperatures.
            Defaults to [].

        calculate_energy (bool, optional): If True, the energy of each
            frame is calculated with the GP. Defaults to False.
        write_model (int, optional): If 0, write never. If 1, write at
            end of run. If 2, write after each training and end of run.
            If 3, write after each time atoms are added and end of run.
            If 4, write after each training and end of run, and back up
            after each write.
        std_tolerance_factor (float, optional): Threshold that determines
            when DFT is called. Specifies a multiple of the current noise
            hyperparameter. If the epistemic uncertainty on a force
            component exceeds this value, DFT is called. Defaults to 1.
        skip (int, optional): Number of frames that are skipped when
            dumping to the output file. Defaults to 0.
        init_atoms (List[int], optional): List of atoms from the input
            structure whose local environments and force components are
            used to train the initial GP model. If None is specified, all
            atoms are used to train the initial GP. Defaults to None.
        output_name (str, optional): Name of the output file. Defaults to
            'otf_run'.
        max_atoms_added (int, optional): Number of atoms added each time
            DFT is called. Defaults to 1.
        freeze_hyps (int, optional): Specifies the number of times the
            hyperparameters of the GP are optimized. After this many
            updates to the GP, the hyperparameters are frozen.
            Defaults to 10.

        n_cpus (int, optional): Number of cpus used during training.
            Defaults to 1.
    """

    def __init__(
        self,
        atoms,
        timestep,
        number_of_steps,
        dft_calc,
        md_engine,
        md_kwargs,
        calculator=None,
        trajectory=None,
        **otf_kwargs,
    ):

        self.structure = FLARE_Atoms.from_ase_atoms(atoms)
        if calculator is not None:
            self.structure.calc = calculator
        self.timestep = timestep
        self.md_engine = md_engine
        self.md_kwargs = md_kwargs
        self._kernels = None

        if md_engine == "VelocityVerlet":
            MD = VelocityVerlet
        elif md_engine == "NVTBerendsen":
            MD = NVTBerendsen
        elif md_engine == "NPTBerendsen":
            MD = NPTBerendsen
        elif md_engine == "NPT":
            MD = NPT_mod
        elif md_engine == "Langevin":
            MD = Langevin
        elif md_engine == "NoseHoover":
            MD = NoseHoover
        else:
            raise NotImplementedError(md_engine + " is not implemented in ASE")

        self.md = MD(
            atoms=self.structure,
            timestep=timestep,
            trajectory=trajectory,
            **md_kwargs,
        )

        force_source = dft_source
        self.flare_calc = self.structure.calc

        # Convert ASE timestep to ps for the output file.
        flare_dt = timestep / (units.fs * 1e3)

        super().__init__(
            dt=flare_dt,
            number_of_steps=number_of_steps,
            gp=self.flare_calc.gp_model,
            force_source=force_source,
            dft_loc=dft_calc,
            dft_input=self.structure,
            **otf_kwargs,
        )

        self.flare_name = self.output_name + "_flare.json"
        self.dft_name = self.output_name + "_dft.pickle"
        self.structure_name = self.output_name + "_atoms.json"
        self.checkpt_files = [
            self.checkpt_name,
            self.flare_name,
            self.dft_name,
            self.structure_name,
            self.dft_xyz,
        ]

    def get_structure_from_input(self, prev_pos_init):
        if prev_pos_init is None:
            self.structure.prev_positions = np.copy(self.structure.positions)
        else:
            assert len(self.structure.positions) == len(
                self.structure.prev_positions
            ), "Previous positions and positions are not same length"
            self.structure.prev_positions = prev_pos_init

    def initialize_train(self):
        super().initialize_train()

        # TODO: Turn this into a "reset" method.
        if not isinstance(self.structure.calc, FLARE_Calculator):
            self.flare_calc.reset()
            self.structure.calc = self.flare_calc

        if self.md_engine == "NPT":
            if not self.md.initialized:
                self.md.initialize()
            else:
                if self.md.have_the_atoms_been_changed():
                    raise NotImplementedError(
                        "You have modified the atoms since the last timestep."
                    )

    def compute_properties(self):
        """
        Compute energies, forces, stresses, and their uncertainties with
            the FLARE ASE calcuator, and write the results to the
            OTF structure object.
        """

        # Change to FLARE calculator if necessary.
        if not isinstance(self.structure.calc, FLARE_Calculator):
            self.flare_calc.reset()
            self.structure.calc = self.flare_calc

        if not self.flare_calc.results:
            self.structure.calc.calculate(self.structure)

    def md_step(self):
        """
        Get new position in molecular dynamics based on the forces predicted by
        FLARE_Calculator or DFT calculator
        """
        # Update previous positions.
        self.structure.prev_positions = np.copy(self.structure.positions)

        # Reset FLARE calculator.
        if self.dft_step:
            self.flare_calc.reset()
            self.structure.calc = self.flare_calc

        # Take MD step.
        # Inside the step() function, get_forces() is called
        self.md.step()
        self.curr_step += 1

    def write_gp(self):
        self.flare_calc.write_model(self.flare_name)

    def rescale_temperature(self, new_pos):
        # call OTF method
        super().rescale_temperature(new_pos)

        # update ASE atoms
        if self.curr_step in self.rescale_steps:
            rescale_ind = self.rescale_steps.index(self.curr_step)
            new_temp = self.rescale_temps[rescale_ind]
            temp_fac = new_temp / self.temperature
            vel_fac = np.sqrt(temp_fac)
            curr_velocities = self.structure.get_velocities()
            self.structure.set_velocities(curr_velocities * vel_fac)

            # Reset thermostat parameters.
            if self.md_engine in ["NVTBerendsen", "NPTBerendsen", "NPT", "Langevin"]:
                self.md.set_temperature(temperature_K=new_temp)
                self.md_kwargs["temperature"] = new_temp * units.kB

    def update_temperature(self):
        self.KE = self.structure.get_kinetic_energy()
        self.temperature = self.structure.get_temperature()

        # Convert velocities to Angstrom / ps.
        self.velocities = self.structure.get_velocities() * units.fs * 1e3

    def update_gp(self, train_atoms, dft_frcs, dft_energy=None, dft_stress=None):
        stds = self.flare_calc.results.get("stds", np.zeros_like(dft_frcs))
        self.output.add_atom_info(train_atoms, stds)

        # Convert ASE stress (xx, yy, zz, yz, xz, xy) to FLARE stress
        # (xx, xy, xz, yy, yz, zz).
        flare_stress = None
        if dft_stress is not None:
            flare_stress = -np.array(
                [
                    dft_stress[0],
                    dft_stress[5],
                    dft_stress[4],
                    dft_stress[1],
                    dft_stress[3],
                    dft_stress[2],
                ]
            )

        if self.force_only:
            dft_energy = None
            flare_stress = None

        # The structure will be added to self.gp.training_structures (struc.Structure).
        # Create a new structure by deepcopy to avoid the forces of the saved
        # structure get modified.
        try:
            struc_to_add = deepcopy(self.structure)
        except TypeError:
            # The structure might be attached with a non-picklable calculator,
            # e.g., when we use LAMMPS empirical potential for training. 
            # When deepcopy fails, create a SinglePointCalculator to store results

            from ase.calculators.singlepoint import SinglePointCalculator

            properties = ["forces", "energy", "stress"]
            results = {
                "forces": self.structure.forces,
                "energy": self.structure.potential_energy,
                "stress": self.structure.stress,
            }

            calc = self.structure.calc
            self.structure.calc = None
            struc_to_add = deepcopy(self.structure)
            struc_to_add.calc = SinglePointCalculator(struc_to_add, **results)
            self.structure.calc = calc

        # update gp model
        self.gp.update_db(
            struc_to_add,
            dft_frcs,
            custom_range=train_atoms,
            energy=dft_energy,
            stress=flare_stress,
        )

        self.gp.set_L_alpha()

        # train model
        if (self.dft_count - 1) < self.freeze_hyps:
            self.train_gp()

        # update mgp model
        if self.flare_calc.use_mapping:
            self.flare_calc.build_map()

        # write model
        if (self.dft_count - 1) < self.freeze_hyps:
            if self.write_model == 2:
                self.write_gp()
        if self.write_model == 3:
            self.write_gp()

    def record_dft_data(self, structure, target_atoms):
        structure.info["target_atoms"] = np.array(target_atoms)
        write(self.dft_xyz, structure, append=True)

    def as_dict(self):

        # DFT module and Trajectory are not picklable
        self.dft_module = self.dft_module.__name__
        md = self.md
        self.md = None
        _kernels = self._kernels
        self._kernels = None
        dft_loc = self.dft_loc
        self.dft_loc = None
        calc = self.dft_input.calc
        self.dft_input.calc = None

        gp = self.gp
        self.gp = None

        # Deepcopy OTF object.
        dct = deepcopy(dict(vars(self)))

        # Reset attributes.
        self.dft_module = eval(self.dft_module)
        self.md = md
        self._kernels = _kernels
        self.dft_loc = dft_loc
        self.dft_input.calc = calc
        self.gp = gp

        # write atoms and flare calculator to separate files
        write(self.structure_name, self.structure)
        dct["atoms"] = self.structure_name

        self.flare_calc.write_model(self.flare_name)
        dct["flare_calc"] = self.flare_name

        # dump dft calculator as pickle
        with open(self.dft_name, "wb") as f:
            pickle.dump(self.dft_loc, f)  # dft_loc is the dft calculator
        dct["dft_loc"] = self.dft_name

        dct["gp"] = self.gp_name

        for key in ["output", "pred_func", "structure", "dft_input", "md"]:
            dct.pop(key)
        dct["md"] = self.md.todict()

        return dct

    @staticmethod
    def from_dict(dct):
        flare_calc_dict = json.load(open(dct["flare_calc"]))

        # Build FLARE_Calculator from dict 
        if flare_calc_dict["class"] == "FLARE_Calculator":
            flare_calc = FLARE_Calculator.from_file(dct["flare_calc"])
            _kernels = None
        # Build SGP_Calculator from dict
        # TODO: we still have the issue that the c++ kernel needs to be 
        # in the current space, otherwise there is Seg Fault
        # That's why there is the _kernels
        elif flare_calc_dict["class"] == "SGP_Calculator":
            from flare_pp.sparse_gp_calculator import SGP_Calculator
            flare_calc, _kernels = SGP_Calculator.from_file(dct["flare_calc"])
        else:
            raise TypeError(
                f"The calculator from {dct['flare_calc']} is not recognized."
            )

        flare_calc.reset()
        dct["atoms"] = read(dct["atoms"])
        dct["calculator"] = flare_calc
        dct.pop("gp")

        with open(dct["dft_loc"], "rb") as f:
            dct["dft_calc"] = pickle.load(f)

        for key in ["dt", "dft_loc"]:
            dct.pop(key)

        new_otf = ASE_OTF(**dct)
        new_otf._kernels = _kernels
        new_otf.dft_count = dct["dft_count"]
        new_otf.curr_step = dct["curr_step"]
        new_otf.std_tolerance = dct["std_tolerance"]

        if new_otf.md_engine == "NPT":
            if not new_otf.md.initialized:
                new_otf.md.initialize()
        return new_otf



"""
any DFT interface can be added here, as long as two functions listed below are implemented

* parse_dft_input(dft_input)
* dft_module.run_dft_par(dft_input, structure, dft_loc, no_cpus)
"""

from flare.dft_interface import qe_util, cp2k_util

dft_software = {"qe": qe_util, "cp2k": cp2k_util}

try:
    import pymatgen
    from flare.dft_interface import vasp_util

    dft_software["vasp"] = vasp_util
except ImportError:
    pass



"""
This module is to provide the same interface as the module `dft_interface`, so we can use ASE atoms and calculators to run OTF
"""
import os
import numpy as np
from copy import deepcopy


def parse_dft_input(atoms):
    pos = np.copy(atoms.positions)
    spc = atoms.get_chemical_symbols()
    cell = np.array(atoms.get_cell())

    # build mass dict
    mass = atoms.get_masses()
    mass_dict = {}
    for i in range(len(spc)):
        spec_ind = str(spc[i])
        if spec_ind not in mass_dict.keys():
            mass_dict[spec_ind] = mass[i]
    return pos, spc, cell, mass_dict


def run_dft_par(atoms, structure, dft_calc, dft_kwargs, **kwargs):
    """
    Assume that the atoms have been updated
    """
    # change from FLARE to DFT calculator
    calc = deepcopy(dft_calc)
    atoms.set_calculator(calc)

    # clean up previous results
    if dft_kwargs is not None and dft_kwargs.get("cleanup", None):
        if "dft" in os.listdir():
            if "WAVECAR" in os.listdir("dft"):
                os.remove("dft/WAVECAR")
        if "WAVECAR" in os.listdir():
            os.remove("WAVECAR")

    # Calculate DFT energy, forces, and stress.
    # Source code for DFT parser:
    # https://wiki.fysik.dtu.dk/ase/_modules/ase/io/espresso.html#read_espresso_out
    # Note that ASE and QE stresses differ by a minus sign.
    forces = atoms.get_forces()
    stress = atoms.get_stress()
    energy = atoms.get_potential_energy()
    # The results will be written to the 'results' dictionary, and structure.forces
    # directly return the atoms.calc.results['forces']

    structure.calc.results = atoms.calc.results

    return forces
